import warnings

import matplotlib.pyplot
import numpy as np
from matplotlib import cm

from amset.plot import amset_base_style, get_figsize, pretty_subplot, styled_plot
from amset.plot.base import BaseMultiTransportPlotter
from amset.plot.transport import (
    fancy_format_doping,
    fancy_format_temp,
    get_lim,
    property_data,
)

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"


class ConvergencePlotter(BaseMultiTransportPlotter):
    def __init__(self, data, use_symbol=False, pad=0.05):
        super().__init__(data)
        self.label_key = "label_symbol" if use_symbol else "label"
        self.pad = pad

    @styled_plot(amset_base_style)
    def get_plot(
        self,
        doping_idx=None,
        temperature_idx=None,
        properties=("conductivity", "seebeck", "thermal conductivity"),
        labels=None,
        x_property=None,
        grid=None,
        height=None,
        width=None,
        doping_type=None,
        conductivity_label=None,
        conductivity_min=None,
        conductivity_max=None,
        log_conductivity=None,
        seebeck_label=None,
        seebeck_min=None,
        seebeck_max=None,
        log_seebeck=None,
        mobility_label=None,
        mobility_min=None,
        mobility_max=None,
        log_mobility=None,
        thermal_conductivity_label=None,
        thermal_conductivity_min=None,
        thermal_conductivity_max=None,
        log_thermal_conductivity=None,
        power_factor_label=None,
        power_factor_min=None,
        power_factor_max=None,
        log_power_factor=None,
        xlabel=None,
        xmin=None,
        xmax=None,
        logx=None,
        legend=True,
        axes=None,
        plt=None,
        style=None,
        no_base_style=False,
        fonts=None,
    ):
        if axes is None:
            if grid is None or len(grid) == 0:
                grid = (1, len(properties))

            if np.product(grid) < len(properties):
                raise ValueError("Grid is not large enough to plot all properties")

            width, height = get_figsize(*grid, width=width, height=height)
            _, axes = pretty_subplot(*grid, width=width, height=height)
        else:
            if not isinstance(axes, (list, tuple, np.ndarray)):
                # single axis given
                grid = (1, 1)
            else:
                axes = np.array(axes)
                if grid is None:
                    grid = axes.shape
                    if len(grid) == 1:
                        grid = (1, grid[0])

            if np.product(grid) < len(properties):
                raise ValueError("Not enough axes to plot all properties")

        if doping_type is not None and doping_idx is not None:
            warnings.warn(
                "Both doping type and doping indexes have been set. This can cause "
                "unexpected behaviour."
            )

        if temperature_idx is None and doping_idx is None:
            if len(self.doping) == 1 or x_property == "temperature":
                warnings.warn(
                    "Either one of temperature-idx or doping-idx should be "
                    "set. Using doping-idx of 0"
                )
                if doping_type:
                    if doping_type == "n":
                        doping_idx = [i for i, d in enumerate(self.doping) if d <= 0]
                    else:
                        doping_idx = [i for i, d in enumerate(self.doping) if d >= 0]
                    if len(doping_idx) == 0:
                        raise ValueError(
                            f"{doping_type}-type doping specified but no doping "
                            f"concentrations of this type are in the calculations"
                        )
                    else:
                        doping_idx = doping_idx[0]
                else:
                    doping_idx = [0]
            else:
                warnings.warn(
                    "Either one of temperature-idx or doping-idx should be "
                    "set. Using temperature-idx of 0"
                )
                temperature_idx = [0]

        prop_dict = {
            "conductivity": {
                "ylabel": conductivity_label,
                "ymin": conductivity_min,
                "ymax": conductivity_max,
                "logy": log_conductivity,
            },
            "mobility": {
                "ylabel": mobility_label,
                "ymin": mobility_min,
                "ymax": mobility_max,
                "logy": log_mobility,
            },
            "seebeck": {
                "ylabel": seebeck_label,
                "ymin": seebeck_min,
                "ymax": seebeck_max,
                "logy": log_seebeck,
            },
            "thermal conductivity": {
                "ylabel": thermal_conductivity_label,
                "ymin": thermal_conductivity_min,
                "ymax": thermal_conductivity_max,
                "logy": log_thermal_conductivity,
            },
            "power factor": {
                "ylabel": power_factor_label,
                "ymin": power_factor_min,
                "ymax": power_factor_max,
                "logy": log_power_factor,
            },
        }

        title = None
        for (x, y), prop in zip(np.ndindex(grid), properties):
            if grid[0] == 1 and grid[1] == 1:
                ax = axes
            elif grid[0] == 1:
                ax = axes[y]
            elif grid[1] == 1:
                ax = axes[x]
            else:
                ax = axes[x, y]

            title = self.plot_property(
                ax,
                prop,
                x_property=x_property,
                doping_type=doping_type,
                temperature_idx=temperature_idx,
                doping_idx=doping_idx,
                xmin=xmin,
                xmax=xmax,
                xlabel=xlabel,
                logx=logx,
                labels=labels,
                **prop_dict[prop],
            )

        if legend:
            if grid[0] == 1 and grid[1] == 1:
                ax = axes
            elif grid[0] == 1:
                ax = axes[-1]
            elif grid[1] == 1:
                ax = axes[0]
            else:
                ax = axes[0, -1]
            ax.legend(loc="upper left", bbox_to_anchor=(1.01, 1), ncol=1, title=title)

        return matplotlib.pyplot

    def plot_property(
        self,
        ax,
        prop,
        x_property=None,
        doping_type=None,
        temperature_idx=None,
        doping_idx=None,
        ylabel=None,
        ymin=None,
        ymax=None,
        logy=None,
        xlabel=None,
        xmin=None,
        xmax=None,
        logx=None,
        labels=None,
    ):
        if temperature_idx is None and doping_idx is None:
            raise ValueError("Either one of temperature-idx or doping-idx must be set")

        if temperature_idx is None:
            temperature_idx = np.arange(self.temperatures.shape[0])
        elif isinstance(temperature_idx, int):
            temperature_idx = [temperature_idx]
        temperatures = self.temperatures[temperature_idx]

        if doping_idx is None:
            doping_idx = np.arange(self.doping.shape[0])
        elif isinstance(doping_idx, int):
            doping_idx = [doping_idx]

        if doping_type == "n":
            doping_idx = [i for i in doping_idx if self.doping[i] <= 0]
        elif doping_type == "p":
            doping_idx = [i for i in doping_idx if self.doping[i] >= 0]
        doping = self.doping[doping_idx]

        if len(doping_idx) != 1 and len(temperature_idx) != 1:
            raise ValueError("A specific temperature or doping index must be chosen")

        data = self._get_data(prop)

        if x_property is None and len(temperatures) == 1:
            x_property = "doping"
        elif x_property is None:
            x_property = "temperature"

        y = np.squeeze(data[:, doping_idx][:, :, temperature_idx])

        if x_property == "doping":
            annotation = fancy_format_temp(temperatures[0])
            x = doping
        elif x_property == "temperature":
            annotation = fancy_format_doping(doping[0])
            x = temperatures
            y = y
        else:
            raise ValueError(f"Unknown x_property: {x_property}")

        if np.any((x > 0) & (x < 0)):
            raise ValueError(
                "Cannot plot n- and p- type properties on the same figure."
                "Use doping-idx to select doping indices"
            )
        x = np.abs(x)

        if labels is None:
            labels = [str(i) for i in range(self.n)]
        elif len(labels) != self.n:
            raise ValueError("Number of labels does not match number of data")

        cmap = cm.get_cmap("viridis_r")
        colors = cmap(np.linspace(0.05, 1, len(y)))

        for yi, label, color in zip(y, labels, colors):
            ax.plot(x, yi, label=label, c=color)

        ylabel = ylabel if ylabel else property_data[prop][self.label_key]
        xlabel = xlabel if xlabel else property_data[x_property][self.label_key]
        logy = logy if logy is not None else property_data[prop]["log"]
        logx = logx if logx is not None else property_data[x_property]["log"]
        ylim = get_lim(y, ymin, ymax, logy, self.pad)
        xlim = get_lim(x, xmin, xmax, logx, self.pad)

        if logy:
            ax.semilogy()
        if logx:
            ax.semilogx()
        ax.set(ylabel=ylabel, xlabel=xlabel, ylim=ylim, xlim=xlim)
        return annotation

    def _get_data(self, prop):
        if prop == "conductivity":
            data = self.conductivity
        elif prop == "seebeck":
            data = self.seebeck
        elif prop == "thermal conductivity":
            data = self.electronic_thermal_conductivity
        elif prop == "mobility":
            data = self.mobility
        elif prop == "power factor":
            data = self.seebeck ** 2 * self.conductivity * 1e-9  # convert to mW/(m K^2)
        else:
            raise ValueError(f"Unrecognised property: {prop}")

        data = np.linalg.eigvalsh(data)
        data = np.average(data, axis=-1)
        return data



import copy
import datetime
import logging
import os
import time
from functools import partial
from pathlib import Path
from typing import Any, Dict, Optional, Union

import numpy as np
from memory_profiler import memory_usage
from monty.json import MSONable
from monty.serialization import loadfn
from pymatgen.core.structure import Structure
from pymatgen.electronic_structure.bandstructure import BandStructure
from pymatgen.electronic_structure.core import Spin
from pymatgen.io.vasp import Vasprun
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.util.string import unicodeify, unicodeify_spacegroup
from tabulate import tabulate

from amset import __version__
from amset.constants import bohr_to_cm, ev_to_hartree, hbar, numeric_types
from amset.core.transport import solve_boltzman_transport_equation
from amset.electronic_structure.common import get_band_structure
from amset.interpolation.bandstructure import Interpolator
from amset.interpolation.projections import ProjectionOverlapCalculator
from amset.interpolation.wavefunction import WavefunctionOverlapCalculator
from amset.io import load_settings, write_settings
from amset.log import initialize_amset_logger, log_banner, log_list
from amset.scattering.calculate import ScatteringCalculator, basic_scatterers
from amset.util import tensor_average, validate_settings

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"

logger = logging.getLogger(__name__)
_kpt_str = "[{k[0]:.2f}, {k[1]:.2f}, {k[2]:.2f}]"


class Runner(MSONable):
    def __init__(
        self,
        band_structure: BandStructure,
        num_electrons: int,
        settings: Dict[str, Any],
    ):
        self._band_structure = band_structure
        self._num_electrons = num_electrons

        # set materials and performance parameters
        # if the user doesn't specify a value then use the default
        self.settings = validate_settings(settings)

    def run(
        self,
        directory: Union[str, Path] = ".",
        return_usage_stats: bool = False,
        prefix: Optional[str] = None,
    ):
        mem_usage, (amset_data, usage_stats) = memory_usage(
            partial(self._run_wrapper, directory=directory, prefix=prefix),
            max_usage=True,
            retval=True,
            interval=0.1,
            include_children=False,
            multiprocess=True,
        )
        log_banner("END")

        logger.info("Timing and memory usage:")
        timing_info = [f"{n} time: {t:.4f} s" for n, t in usage_stats.items()]
        log_list(timing_info + [f"max memory: {mem_usage:.1f} MB"])

        this_date = datetime.datetime.now().strftime("%d %b %Y")
        this_time = datetime.datetime.now().strftime("%H:%M")
        logger.info(f"amset exiting on {this_date} at {this_time}")

        if return_usage_stats:
            usage_stats["max_memory"] = mem_usage
            return amset_data, usage_stats
        else:
            return amset_data

    def _run_wrapper(
        self, directory: Union[str, Path] = ".", prefix: Optional[str] = None
    ):
        if self.settings["print_log"] or self.settings["write_log"]:
            if self.settings["write_log"]:
                log_file = f"{prefix}_amset.log" if prefix else "amset.log"
            else:
                log_file = False

            initialize_amset_logger(
                directory=directory,
                filename=log_file,
                print_log=self.settings["print_log"],
            )

        self._check_wavefunction()

        tt = time.perf_counter()
        _log_amset_intro()
        _log_settings(self)
        _log_structure_information(
            self._band_structure.structure, self.settings["symprec"]
        )
        _log_band_structure_information(self._band_structure)

        amset_data, interpolation_time = self._do_interpolation()
        timing = {"interpolation": interpolation_time}

        amset_data, dos_time = self._do_dos(amset_data)
        timing["dos"] = dos_time

        amset_data, scattering_time = self._do_scattering(amset_data)
        timing["scattering"] = scattering_time

        if isinstance(self.settings["fd_tol"], numeric_types):
            amset_data, timing = self._do_fd_tol(amset_data, directory, prefix, timing)
        else:
            amset_data, timing = self._do_many_fd_tol(
                amset_data, self.settings["fd_tol"], directory, prefix, timing
            )

        timing["total"] = time.perf_counter() - tt
        return amset_data, timing

    def _do_fd_tol(self, amset_data, directory, prefix, timing):
        amset_data.fill_rates_outside_cutoffs()
        amset_data, transport_time = self._do_transport(amset_data)
        timing["transport"] = transport_time

        filepath, writing_time = self._do_writing(amset_data, directory, prefix)
        timing["writing"] = writing_time

        return amset_data, timing

    def _do_many_fd_tol(self, amset_data, fd_tols, directory, prefix, timing):
        prefix = "" if prefix is None else prefix + "_"
        cutoff_pad = _get_cutoff_pad(
            self.settings["pop_frequency"], self.settings["scattering_type"]
        )
        orig_rates = copy.deepcopy(amset_data.scattering_rates)
        mobility_rates_only = self.settings["mobility_rates_only"]

        for fd_tol in sorted(fd_tols)[::-1]:
            # do smallest cutoff last, so the final amset_data is the best result
            for spin in amset_data.spins:
                amset_data.scattering_rates[spin][:] = orig_rates[spin][:]

            amset_data.calculate_fd_cutoffs(
                fd_tol, cutoff_pad=cutoff_pad, mobility_rates_only=mobility_rates_only
            )
            fd_prefix = prefix + f"fd-{fd_tol}"
            _, timing = self._do_fd_tol(amset_data, directory, fd_prefix, timing)
            timing[f"transport ({fd_tol})"] = timing.pop("transport")
            timing[f"writing ({fd_tol})"] = timing.pop("writing")

        return amset_data, timing

    def _check_wavefunction(self):
        if (
            not Path(self.settings["wavefunction_coefficients"]).exists()
            and not self.settings["use_projections"]
        ):
            raise ValueError(
                "Could not find wavefunction coefficients. To run AMSET, the \n"
                "wavefunction coefficients should first be extracted from a WAVECAR \n"
                "file using the 'amset wave' command. See the documentation for more \n"
                "details: https://hackingmaterials.lbl.gov/amset/\n\n"
                "Alternatively, to use the band structure orbital projections to \n"
                "approximate overlap, set 'use_projections' option to true."
            )
        elif self.settings["use_projections"] and not self._band_structure.projections:
            raise ValueError(
                "use_projections is set to true but calculation does not contain\n"
                "orbital projections. Ensure VASP was run with 'LORBIT = 11'\n"
                "Alternatively, use wavefunction coefficients to calculate overlap.\n"
                "Wavefunction coefficients can be extracted from a VASP WAVECAR\n"
                "file using the 'amset wave' command. See the documentation for more\n"
                "details: https://hackingmaterials.lbl.gov/amset/\n\n"
            )
        elif self.settings["use_projections"]:
            logger.info(
                "Using orbital projections to approximate wavefunction overlap. This "
                "can result in inaccurate results. I hope you know what you are doing."
            )

    def _do_interpolation(self):
        log_banner("INTERPOLATION")
        t0 = time.perf_counter()

        interpolater = Interpolator(
            self._band_structure,
            num_electrons=self._num_electrons,
            interpolation_factor=self.settings["interpolation_factor"],
            soc=self.settings["soc"],
        )

        amset_data = interpolater.get_amset_data(
            energy_cutoff=self.settings["energy_cutoff"],
            scissor=self.settings["scissor"],
            bandgap=self.settings["bandgap"],
            symprec=self.settings["symprec"],
            nworkers=self.settings["nworkers"],
        )

        if set(self.settings["scattering_type"]).issubset(set(basic_scatterers)):
            overlap_calculator = None
        elif self.settings["use_projections"]:
            overlap_calculator = ProjectionOverlapCalculator.from_band_structure(
                self._band_structure,
                energy_cutoff=self.settings["energy_cutoff"],
                symprec=self.settings["symprec"],
            )
        else:
            overlap_calculator = WavefunctionOverlapCalculator.from_file(
                self.settings["wavefunction_coefficients"]
            )
        amset_data.set_overlap_calculator(overlap_calculator)

        return amset_data, time.perf_counter() - t0

    def _do_dos(self, amset_data):
        log_banner("DOS")
        t0 = time.perf_counter()

        amset_data.calculate_dos(
            estep=self.settings["dos_estep"], progress_bar=self.settings["print_log"]
        )
        amset_data.set_doping_and_temperatures(
            self.settings["doping"], self.settings["temperatures"]
        )

        cutoff_pad = _get_cutoff_pad(
            self.settings["pop_frequency"], self.settings["scattering_type"]
        )

        if isinstance(self.settings["fd_tol"], numeric_types):
            fd_tol = self.settings["fd_tol"]
        else:
            fd_tol = min(self.settings["fd_tol"])

        mob_only = self.settings["mobility_rates_only"]
        amset_data.calculate_fd_cutoffs(
            fd_tol, cutoff_pad=cutoff_pad, mobility_rates_only=mob_only
        )
        return amset_data, time.perf_counter() - t0

    def _do_scattering(self, amset_data):
        log_banner("SCATTERING")
        t0 = time.perf_counter()

        cutoff_pad = _get_cutoff_pad(
            self.settings["pop_frequency"], self.settings["scattering_type"]
        )

        scatter = ScatteringCalculator(
            self.settings,
            amset_data,
            cutoff_pad,
            scattering_type=self.settings["scattering_type"],
            progress_bar=self.settings["print_log"],
            cache_wavefunction=self.settings["cache_wavefunction"],
            nworkers=self.settings["nworkers"],
        )

        amset_data.set_scattering_rates(
            scatter.calculate_scattering_rates(), scatter.scatterer_labels
        )
        return amset_data, time.perf_counter() - t0

    def _do_transport(self, amset_data):
        log_banner("TRANSPORT")
        t0 = time.perf_counter()
        transport_properties = solve_boltzman_transport_equation(
            amset_data,
            separate_mobility=self.settings["separate_mobility"],
            calculate_mobility=self.settings["calculate_mobility"],
            progress_bar=self.settings["print_log"],
        )
        amset_data.set_transport_properties(*transport_properties)
        return amset_data, time.perf_counter() - t0

    def _do_writing(self, amset_data, directory, prefix):
        log_banner("RESULTS")
        _log_results_summary(amset_data, self.settings)

        abs_dir = os.path.abspath(directory)
        t0 = time.perf_counter()

        if not os.path.exists(abs_dir):
            os.makedirs(abs_dir)

        if self.settings["write_input"]:
            self.write_settings(abs_dir)

        if self.settings["file_format"] is None:
            full_filename = None
        else:
            filename = amset_data.to_file(
                directory=abs_dir,
                write_mesh_file=self.settings["write_mesh"],
                prefix=prefix,
                file_format=self.settings["file_format"],
            )

            if isinstance(filename, tuple):
                full_filename = "\nand\n".join(
                    [str(Path(abs_dir) / f) for f in filename]
                )
            else:
                full_filename = Path(abs_dir) / filename

            logger.info(f"Results written to:\n{full_filename}")
        return full_filename, time.perf_counter() - t0

    @staticmethod
    def from_vasprun(
        vasprun: Union[str, Path, Vasprun], settings: Dict[str, Any]
    ) -> "Runner":
        """Initialise an AmsetRunner from a Vasprun.

        The nelect and soc options will be determined from the Vasprun
        automatically.

        Args:
            vasprun: Path to a vasprun or a Vasprun pymatgen object.
            settings: AMSET settings.

        Returns:
            A Runner object that can be used to calculate transport properties.
        """
        if not isinstance(vasprun, Vasprun):
            vasprun = Vasprun(vasprun, parse_projected_eigen=True)

        zwk_option = settings.get("zero_weighted_kpoints", "keep")
        band_structure = get_band_structure(vasprun, zero_weighted=zwk_option)
        nelect = vasprun.parameters["NELECT"]
        settings["soc"] = vasprun.parameters["LSORBIT"]

        return Runner(band_structure, nelect, settings)

    @staticmethod
    def from_directory(
        directory: Union[str, Path] = ".",
        input_file: Optional[Union[str, Path]] = None,
        settings_file: Optional[Union[str, Path]] = None,
        settings_override: Optional[Dict[str, Any]] = None,
    ):
        """
        Initialize amset Runner from a directory.

        If input_file or settings_file are not specified, the code will look in the
        specified directory for these files.

        Args:
            directory: A directory.
            input_file: An input file path, can either be vasprun.xml(.gz) or a
                band_structure_data.json(.gz) file containing the keys:

                - "nelect" (int): The number of electrons in the system.
                - "band_structure (BandStructure)": A pymatgen band structure object.
            settings_file: Path to settings file.
            settings_override: Settings that will be used to override the settings
                in the settings file.

        Returns:
            A Runner instance that can be used to run amset.
        """
        directory = Path(directory)

        if not settings_file:
            settings_file = directory / "settings.yaml"
        settings = load_settings(settings_file)

        if settings_override:
            settings.update(settings_override)

        run_type, input_file = _get_run_type(directory, input_file)
        if run_type == "vasprun":
            return Runner.from_vasprun(input_file, settings)
        elif run_type == "band_structure":
            data = loadfn(input_file)
            nelect = data["nelect"]
            band_structure = data["band_structure"]
            return Runner(band_structure, nelect, settings)
        else:
            raise ValueError(f"Unrecognised run type: {run_type}")

    def write_settings(self, directory: str = ".", prefix: Optional[str] = None):
        prefix = "" if prefix is None else f"{prefix}_"
        filename = Path(directory) / f"{prefix}amset_settings.yaml"
        write_settings(self.settings, filename)


def _log_amset_intro():
    now = datetime.datetime.now()
    logger.info(
        """
               █████╗ ███╗   ███╗███████╗███████╗████████╗
              ██╔══██╗████╗ ████║██╔════╝██╔════╝╚══██╔══╝
              ███████║██╔████╔██║███████╗█████╗     ██║
              ██╔══██║██║╚██╔╝██║╚════██║██╔══╝     ██║
              ██║  ██║██║ ╚═╝ ██║███████║███████╗   ██║
              ╚═╝  ╚═╝╚═╝     ╚═╝╚══════╝╚══════╝   ╚═╝

                                                  v{}

    Ganose, A. M., Park, J., Faghaninia, A., Woods-Robinson,
    R., Persson, K. A., Jain, A. Efficient calculation of
    carrier scattering rates from first principles.
    Nat. Commun. 12, 2222 (2021)


amset starting on {} at {}""".format(
            __version__, now.strftime("%d %b %Y"), now.strftime("%H:%M")
        )
    )


def _log_structure_information(structure: Structure, symprec):
    log_banner("STRUCTURE")
    logger.info("Structure information:")

    comp = structure.composition
    lattice = structure.lattice
    formula = comp.get_reduced_formula_and_factor(iupac_ordering=True)[0]

    if not symprec:
        symprec = 1e-32

    sga = SpacegroupAnalyzer(structure, symprec=symprec)
    spg = unicodeify_spacegroup(sga.get_space_group_symbol())

    comp_info = [
        f"formula: {unicodeify(formula)}",
        f"# sites: {structure.num_sites}",
        f"space group: {spg}",
    ]
    log_list(comp_info)

    logger.info("Lattice:")
    lattice_info = [
        "a, b, c [Å]: {:.2f}, {:.2f}, {:.2f}".format(*lattice.abc),
        "α, β, γ [°]: {:.0f}, {:.0f}, {:.0f}".format(*lattice.angles),
    ]
    log_list(lattice_info)


_tensor_str = """
    │   [[{:6.2f} {:6.2f} {:6.2f}]
    │    [{:6.2f} {:6.2f} {:6.2f}]
    │    [{:6.2f} {:6.2f} {:6.2f}]]"""

_elastic_tensor_str = """
    │   [[{:6.1f} {:6.1f} {:6.1f} {:6.1f} {:6.1f} {:6.1f}]
    │    [{:6.1f} {:6.1f} {:6.1f} {:6.1f} {:6.1f} {:6.1f}]
    │    [{:6.1f} {:6.1f} {:6.1f} {:6.1f} {:6.1f} {:6.1f}]
    │    [{:6.1f} {:6.1f} {:6.1f} {:6.1f} {:6.1f} {:6.1f}]
    │    [{:6.1f} {:6.1f} {:6.1f} {:6.1f} {:6.1f} {:6.1f}]
    │    [{:6.1f} {:6.1f} {:6.1f} {:6.1f} {:6.1f} {:6.1f}]]"""

_piezo_tensor_str = """
    │   [[{:7.4f} {:7.4f} {:7.4f} {:7.4f} {:7.4f} {:7.4f}]
    │    [{:7.4f} {:7.4f} {:7.4f} {:7.4f} {:7.4f} {:7.4f}]
    │    [{:7.4f} {:7.4f} {:7.4f} {:7.4f} {:7.4f} {:7.4f}]]"""


def _log_settings(runner: Runner):
    from pymatgen.core.tensors import Tensor

    def ff(prop):
        # format tensor properties
        if isinstance(prop, np.ndarray):
            if prop.shape == (3, 3):
                return _tensor_str.format(*prop.ravel())
            elif prop.shape == (3, 3, 3):
                return _piezo_tensor_str.format(*Tensor(prop).voigt.ravel())
            elif prop.shape == (3, 3, 3, 3):
                return _elastic_tensor_str.format(*Tensor(prop).voigt.ravel())

        return prop

    log_banner("SETTINGS")
    logger.info("Run parameters:")
    p = [f"{k}: {ff(v)}" for k, v in runner.settings.items() if v is not None]
    log_list(p)


def _log_band_structure_information(band_structure: BandStructure):
    log_banner("BAND STRUCTURE")

    info = [
        f"# bands: {band_structure.nb_bands}",
        f"# k-points: {len(band_structure.kpoints)}",
        f"Fermi level: {band_structure.efermi:.3f} eV",
        f"spin polarized: {band_structure.is_spin_polarized}",
        f"metallic: {band_structure.is_metal()}",
    ]
    logger.info("Input band structure information:")
    log_list(info)

    if band_structure.is_metal():
        return

    logger.info("Band gap:")
    band_gap_info = []

    bg_data = band_structure.get_band_gap()
    if not bg_data["direct"]:
        band_gap_info.append("indirect band gap: {:.3f} eV".format(bg_data["energy"]))

    direct_data = band_structure.get_direct_band_gap_dict()
    direct_bg = min(spin_data["value"] for spin_data in direct_data.values())
    band_gap_info.append(f"direct band gap: {direct_bg:.3f} eV")

    direct_kpoint = []
    for spin, spin_data in direct_data.items():
        direct_kindex = spin_data["kpoint_index"]
        kpt_str = _kpt_str.format(k=band_structure.kpoints[direct_kindex].frac_coords)
        direct_kpoint.append(kpt_str)

    band_gap_info.append("direct k-point: {}".format(", ".join(direct_kpoint)))
    log_list(band_gap_info)

    vbm_data = band_structure.get_vbm()
    cbm_data = band_structure.get_cbm()

    logger.info("Valence band maximum:")
    _log_band_edge_information(band_structure, vbm_data)

    logger.info("Conduction band minimum:")
    _log_band_edge_information(band_structure, cbm_data)


def _log_band_edge_information(band_structure, edge_data):
    """Log data about the valence band maximum or conduction band minimum.

    Args:
        band_structure: A band structure.
        edge_data (dict): The :obj:`dict` from ``bs.get_vbm()`` or
            ``bs.get_cbm()``
    """
    if band_structure.is_spin_polarized:
        spins = edge_data["band_index"].keys()
        b_indices = [
            ", ".join([str(i + 1) for i in edge_data["band_index"][spin]])
            + f"({spin.name.capitalize()})"
            for spin in spins
        ]
        b_indices = ", ".join(b_indices)
    else:
        b_indices = ", ".join([str(i + 1) for i in edge_data["band_index"][Spin.up]])

    kpoint = edge_data["kpoint"]
    kpoint_str = _kpt_str.format(k=kpoint.frac_coords)

    info = [
        "energy: {:.3f} eV".format(edge_data["energy"]),
        f"k-point: {kpoint_str}",
        f"band indices: {b_indices}",
    ]
    log_list(info)


def _log_results_summary(amset_data, output_parameters):
    results_summary = []

    doping = [d * (1 / bohr_to_cm) ** 3 for d in amset_data.doping]
    temps = amset_data.temperatures

    if output_parameters["calculate_mobility"] and not amset_data.is_metal:
        logger.info(
            "Average conductivity (σ), Seebeck (S) and mobility (μ)" " results:"
        )
        headers = ("conc [cm⁻³]", "temp [K]", "σ [S/m]", "S [µV/K]", "μ [cm²/Vs]")
        for c, t in np.ndindex(amset_data.fermi_levels.shape):
            results = (
                doping[c],
                temps[t],
                tensor_average(amset_data.conductivity[c, t]),
                tensor_average(amset_data.seebeck[c, t]),
                tensor_average(amset_data.mobility["overall"][c, t]),
            )
            results_summary.append(results)

    else:
        logger.info("Average conductivity (σ) and Seebeck (S) results:")
        headers = ("conc [cm⁻³]", "temp [K]", "σ [S/m]", "S [µV/K]")
        for c, t in np.ndindex(amset_data.fermi_levels.shape):
            results = (
                doping[c],
                temps[t],
                tensor_average(amset_data.conductivity[c, t]),
                tensor_average(amset_data.seebeck[c, t]),
            )
            results_summary.append(results)

    table = tabulate(
        results_summary,
        headers=headers,
        numalign="right",
        stralign="center",
        floatfmt=(".2e", ".1f", ".2e", ".2e", ".1f"),
    )
    logger.info(table)

    if output_parameters["separate_mobility"] and not amset_data.is_metal:
        labels = amset_data.scattering_labels
        logger.info("Mobility breakdown by scattering mechanism, in cm²/Vs:")
        headers = ["conc [cm⁻³]", "temp [K]"] + labels

        results_summary = []
        for c, t in np.ndindex(amset_data.fermi_levels.shape):
            results = [doping[c], temps[t]]
            results += [tensor_average(amset_data.mobility[s][c, t]) for s in labels]
            results_summary.append(results)

        table = tabulate(
            results_summary,
            headers=headers,
            numalign="right",
            stralign="center",
            floatfmt=[".2e", ".1f"] + [".2e"] * len(labels),
        )
        logger.info(table)


def _get_cutoff_pad(pop_frequency, scattering_type):
    cutoff_pad = 0
    if pop_frequency and ("POP" in scattering_type or scattering_type == "auto"):
        # convert from THz to angular frequency in Hz
        pop_frequency = pop_frequency * 1e12 * 2 * np.pi

        # use the phonon energy to pad the fermi dirac cutoffs, this is because
        # pop scattering from a kpoints, k, to kpoints with energies above and below
        # k. We therefore need k-points above and below to be within the cut-offs
        # otherwise scattering cannot occur
        cutoff_pad = pop_frequency * hbar * ev_to_hartree
    return cutoff_pad


def _get_run_type(directory: Path, input_file: Optional[str]):
    if input_file is None:
        vr_files = list(directory.glob("*vasprun*"))
        bs_files = list(directory.glob("*band_structure_data*"))
        if len(vr_files) > 0:
            input_file = vr_files[0]
        elif len(bs_files) > 0:
            input_file = bs_files[0]
        else:
            raise ValueError(
                "Could not find vasprun.xml, vasprun.xml.gz or band_structure_data.json"
                f" in {directory}"
            )

    input_file = str(input_file)
    if "xml" in input_file:
        run_type = "vasprun"
    elif "json" in input_file:
        run_type = "band_structure"
    else:
        raise ValueError(f"Could not determine input file type: {input_file}")

    return run_type, input_file



import itertools

import numpy as np
from monty.dev import requires
from pymatgen.io.vasp import Potcar, Vasprun
from tqdm.auto import tqdm

from amset.constants import int_to_spin, numeric_types, spin_to_int
from amset.wavefunction.common import sample_random_kpoints

try:
    import pawpyseed as pawpy
except ImportError:
    pawpy = None

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"

pawpy_msg = (
    "Pawpyseed is required for extracting wavefunction coefficients\nFollow the"
    "installation instructions at https://github.com/kylebystrom/pawpyseed"
)


@requires(pawpy, pawpy_msg)
def get_wavefunction(
    potcar="POTCAR", wavecar="WAVECAR", vasprun="vasprun.xml", directory=None
):
    from pawpyseed.core import pawpyc
    from pawpyseed.core.wavefunction import CoreRegion, Wavefunction

    if directory:
        wf = Wavefunction.from_directory(path=directory)
    else:
        if isinstance(vasprun, str):
            vasprun = Vasprun(vasprun)

        if isinstance(potcar, str):
            potcar = Potcar.from_file(potcar)

        ngx = vasprun.parameters["NGX"]
        ngy = vasprun.parameters["NGY"]
        ngz = vasprun.parameters["NGZ"]
        dim = np.array([ngx, ngy, ngz])
        symprec = vasprun.parameters["SYMPREC"]
        structure = vasprun.final_structure

        pwf = pawpyc.PWFPointer(wavecar, vasprun)
        core_region = CoreRegion(potcar)

        wf = Wavefunction(structure, pwf, core_region, dim, symprec, False)

    wf = wf.desymmetrized_copy(time_reversal_symmetry=False, symprec=1e-4)
    return wf


@requires(pawpy, pawpy_msg)
def get_wavefunction_coefficients(
    wavefunction, bs, iband=None, ikpoints=None, encut=600, pbar=True
):
    from pawpyseed.core.momentum import MomentumMatrix

    mm = MomentumMatrix(wavefunction, encut=encut)
    if not iband:
        iband = {}

    if not ikpoints:
        ikpoints = np.arange(mm.wf.kpts.shape[0])

    coeffs = {}
    for spin_idx in range(wavefunction.nspin):
        spin = int_to_spin[spin_idx]
        spin_iband = iband.get(spin, None)

        coeffs[spin] = _get_spin_wavefunction_coefficients(
            mm, bs, spin, iband=spin_iband, ikpoints=ikpoints, pbar=pbar
        )

    return coeffs, mm.momentum_grid, wavefunction.kpts[ikpoints]


def _get_spin_wavefunction_coefficients(
    mm, bs, spin, iband=None, ikpoints=None, pbar=True
):
    from amset.constants import output_width

    if iband is None:
        iband = np.arange(bs.bands[spin].shape[0], dtype=int)

    elif isinstance(iband, numeric_types):
        iband = [iband]

    if ikpoints is None:
        ikpoints = np.arange(mm.wf.kpts.shape[0])

    ncoeffs = mm.momentum_grid.shape[0]
    ns = spin_to_int[spin]
    coeffs = np.zeros((len(iband), len(ikpoints), ncoeffs), dtype=complex)

    state_idxs = list(itertools.product(enumerate(iband), enumerate(ikpoints)))
    if pbar:
        state_idxs = tqdm(state_idxs, ncols=output_width)

    for (i, nb), (j, nk) in state_idxs:
        coeffs[i, j] = mm.get_reciprocal_fullfw(nb, nk, ns)
        coeffs[i, j] /= np.linalg.norm(coeffs[i, j])

    return coeffs


@requires(pawpy, pawpy_msg)
def get_converged_encut(
    wavefunction,
    bs,
    iband=None,
    ikpoints=None,
    max_encut=500,
    n_samples=1000,
    std_tol=0.02,
):
    from pawpyseed.core.momentum import MomentumMatrix

    nspins = wavefunction.nspin
    if ikpoints is None:
        ikpoints = np.arange(wavefunction.kpts.shape[0])

    if iband is None:
        iband = {s: len(bs.bands[s]) for s in bs.spins}

    sample_points = sample_random_kpoints(nspins, ikpoints, iband, n_samples)
    origin = sample_points[0]
    sample_points = sample_points[1:]

    mm = MomentumMatrix(wavefunction, encut=max_encut)
    true_overlaps = get_overlaps(mm, origin, sample_points)

    # filter points to only include these with reasonable overlaps
    mask = true_overlaps > 0.05
    true_overlaps = true_overlaps[mask]
    sample_points = sample_points[mask]

    for encut in np.arange(100, max_encut, 50):
        mm = MomentumMatrix(wavefunction, encut=encut)
        fake_overlaps = get_overlaps(mm, origin, sample_points)
        diff = (true_overlaps / fake_overlaps) - 1
        if diff.std() < std_tol:
            return encut

    return max_encut


def get_overlaps(mm, origin, points):
    coeffs = np.zeros((len(points), mm.momentum_grid.shape[0]), dtype=complex)

    origin = mm.get_reciprocal_fullfw(origin[1], origin[2], origin[0])
    origin /= np.linalg.norm(origin)

    for i, (s_idx, b_idx, k_idx) in enumerate(points):
        coeffs[i] = mm.get_reciprocal_fullfw(b_idx, k_idx, s_idx)
        coeffs[i] /= np.linalg.norm(coeffs[i])

    overlaps = np.abs((np.conj(origin[:, None]) * coeffs.T).sum(axis=0)) ** 2
    return overlaps



import logging
import time
from collections import defaultdict
from os.path import join as joinpath
from typing import Dict, List, Optional

import numpy as np
from monty.json import MSONable
from monty.serialization import dumpfn
from pymatgen.core.structure import Structure
from pymatgen.electronic_structure.core import Spin
from tabulate import tabulate

from amset.constants import bohr_to_cm, boltzmann_au, cm_to_bohr
from amset.constants import defaults as defaults
from amset.constants import ev_to_hartree, hartree_to_ev, spin_name
from amset.electronic_structure.common import (
    get_angstrom_structure,
    get_cbm_energy,
    get_vbm_energy,
)
from amset.electronic_structure.dos import FermiDos
from amset.electronic_structure.fd import dfdde
from amset.electronic_structure.tetrahedron import TetrahedralBandStructure
from amset.interpolation.momentum import MRTACalculator
from amset.io import write_mesh
from amset.log import log_list, log_time_taken
from amset.util import cast_dict_list, groupby, tensor_average

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"

logger = logging.getLogger(__name__)
_kpt_str = "[{k[0]:.5f} {k[1]:.5f} {k[2]:.5f}]"


class AmsetData(MSONable):
    def __init__(
        self,
        structure: Structure,
        energies: Dict[Spin, np.ndarray],
        vvelocities_product: Dict[Spin, np.ndarray],
        velocities: Dict[Spin, np.ndarray],
        kpoint_mesh: np.ndarray,
        kpoints: np.ndarray,
        ir_kpoints_idx: np.ndarray,
        ir_to_full_kpoint_mapping: np.ndarray,
        tetrahedra: np.ndarray,
        ir_tetrahedra_info: np.ndarray,
        efermi: float,
        num_electrons: float,
        is_metal: bool,
        soc: bool,
        vb_idx: Optional[Dict[Spin, int]] = None,
    ):
        self.structure = structure
        self.velocities_product = vvelocities_product
        self.kpoint_mesh = kpoint_mesh
        self.intrinsic_fermi_level = efermi
        self.ir_to_full_kpoint_mapping = ir_to_full_kpoint_mapping
        self._soc = soc
        self.num_electrons = num_electrons
        self.is_metal = is_metal
        self.vb_idx = vb_idx
        self.spins = list(energies.keys())
        self.velocities = {s: v.transpose((0, 2, 1)) for s, v in velocities.items()}

        self.dos = None
        self.scattering_rates = None
        self.scattering_labels = None
        self.doping = None
        self.temperatures = None
        self.fermi_levels = None
        self.electron_conc = None
        self.hole_conc = None
        self.conductivity = None
        self.seebeck = None
        self.electronic_thermal_conductivity = None
        self.mobility = None
        self.overlap_calculator = None
        self.mrta_calculator = None
        self.fd_cutoffs = None

        self.grouped_ir_to_full = groupby(
            np.arange(len(kpoints)), ir_to_full_kpoint_mapping
        )

        self.tetrahedral_band_structure = TetrahedralBandStructure.from_data(
            energies,
            kpoints,
            tetrahedra,
            structure,
            ir_kpoints_idx,
            ir_to_full_kpoint_mapping,
            *ir_tetrahedra_info,
        )

        logger.info("Initializing momentum relaxation time factor calculator")
        self.mrta_calculator = MRTACalculator.from_data(kpoints, self.velocities)

    @property
    def energies(self):
        return self.tetrahedral_band_structure.energies

    @property
    def kpoints(self):
        return self.tetrahedral_band_structure.kpoints

    @property
    def ir_kpoints(self):
        return self.kpoints[self.tetrahedral_band_structure.ir_kpoints_idx]

    @property
    def ir_kpoints_idx(self):
        return self.tetrahedral_band_structure.ir_kpoints_idx

    def set_overlap_calculator(self, overlap_calculator):
        if overlap_calculator is not None:
            equal = check_nbands_equal(overlap_calculator, self)
            if not equal:
                raise RuntimeError(
                    "Overlap calculator does not have the correct number of bands\n"
                    "If using wavefunction coefficients, ensure they were generated "
                    "using the same energy_cutoff (not encut)"
                )

        self.overlap_calculator = overlap_calculator

    def calculate_dos(
        self,
        estep: float = defaults["dos_estep"],
        progress_bar: bool = defaults["print_log"],
    ):
        """
        Args:
            estep: The DOS energy step in eV, where smaller numbers give more
                accuracy but are more expensive.
            progress_bar: Show a progress bar for DOS calculation.
        """
        emin = np.min([np.min(spin_eners) for spin_eners in self.energies.values()])
        emax = np.max([np.max(spin_eners) for spin_eners in self.energies.values()])
        epoints = int(round((emax - emin) / (estep * ev_to_hartree)))
        energies = np.linspace(emin, emax, epoints)
        dos_weight = 1 if self._soc or len(self.spins) == 2 else 2

        logger.info("DOS parameters:")
        log_list(
            [
                f"emin: {emin * hartree_to_ev:.2f} eV",
                f"emax: {emax * hartree_to_ev:.2f} eV",
                f"dos weight: {dos_weight}",
                f"n points: {epoints}",
            ]
        )

        logger.info("Generating tetrahedral DOS:")
        t0 = time.perf_counter()
        emesh, dos = self.tetrahedral_band_structure.get_density_of_states(
            energies=energies, progress_bar=progress_bar
        )
        log_time_taken(t0)

        num_electrons = self.num_electrons if self.is_metal else None

        self.dos = FermiDos(
            self.intrinsic_fermi_level,
            emesh,
            dos,
            self.structure,
            atomic_units=True,
            dos_weight=dos_weight,
            num_electrons=num_electrons,
        )

    def set_doping_and_temperatures(self, doping: np.ndarray, temperatures: np.ndarray):
        if not self.dos:
            raise RuntimeError(
                "The DOS should be calculated (AmsetData.calculate_dos) before "
                "setting doping levels."
            )

        if doping is None:
            # Generally this is for metallic systems; here we use the intrinsic Fermi
            # level
            self.doping = [0]
            print("doping is none")
        else:
            self.doping = doping * (1 / cm_to_bohr) ** 3

        self.temperatures = temperatures

        self.fermi_levels = np.zeros((len(doping), len(temperatures)))
        self.electron_conc = np.zeros((len(doping), len(temperatures)))
        self.hole_conc = np.zeros((len(doping), len(temperatures)))

        fermi_level_info = []
        tols = np.logspace(-5, 0, 6)
        for n, t in np.ndindex(self.fermi_levels.shape):
            for i, tol in enumerate(tols):
                # Finding the Fermi level is quite fickle. Enumerate multiple
                # tolerances and use the first one that works!
                try:
                    if self.doping[n] == 0:
                        self.fermi_levels[n, t] = self.dos.get_fermi_from_num_electrons(
                            self.num_electrons,
                            temperatures[t],
                            tol=tol / 1000,
                            precision=10,
                        )
                    else:
                        (
                            self.fermi_levels[n, t],
                            self.electron_conc[n, t],
                            self.hole_conc[n, t],
                        ) = self.dos.get_fermi(
                            self.doping[n],
                            temperatures[t],
                            tol=tol,
                            precision=10,
                            return_electron_hole_conc=True,
                        )
                    break
                except ValueError:
                    if i == len(tols) - 1:
                        raise ValueError(
                            "Could not calculate Fermi level position."
                            "Try a denser k-point mesh."
                        )
                    else:
                        pass

            fermi_level_info.append(
                (doping[n], temperatures[t], self.fermi_levels[n, t] * hartree_to_ev)
            )

        table = tabulate(
            fermi_level_info,
            headers=("conc [cm⁻³]", "temp [K]", "E_fermi [eV]"),
            numalign="right",
            stralign="center",
            floatfmt=(".2e", ".1f", ".4f"),
        )
        logger.info("Calculated Fermi levels:")
        logger.info(table)

    def calculate_fd_cutoffs(
        self,
        fd_tolerance: Optional[float] = 0.01,
        cutoff_pad: float = 0.0,
        max_moment: int = 2,
        mobility_rates_only: bool = False,
    ):
        energies = self.dos.energies
        vv = {s: v.transpose((0, 3, 1, 2)) for s, v in self.velocities_product.items()}
        _, vvdos = self.tetrahedral_band_structure.get_density_of_states(
            energies, integrand=vv, sum_spins=True, use_cached_weights=True
        )
        vvdos = tensor_average(vvdos)
        # vvdos = np.array(self.dos.get_densities())

        # three fermi integrals govern transport properties:
        #   1. df/de controls conductivity and mobility
        #   2. (e-u) * df/de controls Seebeck
        #   3. (e-u)^2 df/de controls electronic thermal conductivity
        # take the absolute sum of the integrals across all doping and
        # temperatures. this gives us the energies that are important for
        # transport
        if fd_tolerance:

            def get_min_max_cutoff(cumsum):
                min_idx = np.where(cumsum < fd_tolerance / 2)[0].max()
                max_idx = np.where(cumsum > (1 - fd_tolerance / 2))[0].min()
                return energies[min_idx], energies[max_idx]

            min_cutoff = np.inf
            max_cutoff = -np.inf
            for n, t in np.ndindex(self.fermi_levels.shape):
                ef = self.fermi_levels[n, t]
                temp = self.temperatures[t]
                dfde = -dfdde(energies, ef, temp * boltzmann_au)

                for moment in range(max_moment + 1):
                    weight = np.abs((energies - ef) ** moment * dfde)
                    weight_dos = weight * vvdos
                    weight_cumsum = np.cumsum(weight_dos)
                    weight_cumsum /= np.max(weight_cumsum)

                    cmin, cmax = get_min_max_cutoff(weight_cumsum)
                    min_cutoff = min(cmin, min_cutoff)
                    max_cutoff = max(cmax, max_cutoff)

                    # import matplotlib.pyplot as plt
                    # ax = plt.gca()
                    # plt.plot(energies / units.eV, weight / weight.max())
                    # plt.plot(energies / units.eV, vvdos / vvdos.max())
                    # plt.plot(energies / units.eV, weight_dos / weight_dos.max())
                    # plt.plot(energies / units.eV, weight_cumsum / weight_cumsum.max())
                    # ax.set(xlim=(4, 7.5))
                    # plt.show()

        else:
            min_cutoff = energies.min()
            max_cutoff = energies.max()

        if mobility_rates_only:
            vbm = get_vbm_energy(self.energies, self.vb_idx)
            cbm = get_cbm_energy(self.energies, self.vb_idx)
            mid_gap = (cbm + vbm) / 2
            if np.all(self.doping < 0):
                # only electron mobility so don't calculate valence band rates
                min_cutoff = max(min_cutoff, mid_gap)
            elif np.all(self.doping < 0):
                # only hole mobility so don't calculate conudction band rates
                max_cutoff = min(max_cutoff, mid_gap)

        min_cutoff -= cutoff_pad
        max_cutoff += cutoff_pad

        logger.info("Calculated Fermi–Dirac cut-offs:")
        log_list(
            [
                f"min: {min_cutoff * hartree_to_ev:.3f} eV",
                f"max: {max_cutoff * hartree_to_ev:.3f} eV",
            ]
        )
        self.fd_cutoffs = (min_cutoff, max_cutoff)

    def set_scattering_rates(
        self, scattering_rates: Dict[Spin, np.ndarray], scattering_labels: List[str]
    ):
        for spin in self.spins:
            s = (len(self.doping), len(self.temperatures)) + self.energies[spin].shape
            if scattering_rates[spin].shape[1:] != s:
                raise ValueError(
                    "Shape of scattering_type rates array does not match the "
                    "number of dopings, temperatures, bands, or kpoints"
                )

            if scattering_rates[spin].shape[0] != len(scattering_labels):
                raise ValueError(
                    "Number of scattering_type rates does not match number of "
                    "scattering_type labels"
                )

        self.scattering_rates = scattering_rates
        self.scattering_labels = scattering_labels

    def fill_rates_outside_cutoffs(self, fill_value=None):
        if self.scattering_rates is None:
            raise ValueError("Scattering rates must be set before being filled")

        min_fd, max_fd = self.fd_cutoffs
        snt_fill = fill_value
        for spin, spin_energies in self.energies.items():
            mask = (spin_energies < min_fd) | (spin_energies > max_fd)
            any_in_mask = np.any(~mask)  # sometimes no rates calculated for spin
            rate_info = defaultdict(list)
            for s, n, t in np.ndindex(self.scattering_rates[spin].shape[:3]):
                if fill_value is None and any_in_mask:
                    # get average log rate inside cutoffs
                    snt_fill = np.log(self.scattering_rates[spin][s, n, t, ~mask])
                    snt_fill = np.exp(snt_fill.mean())
                elif fill_value is None:
                    # no rates have been calculated for this spin channel so cannot
                    # average them. In this rare case we use a rate of 10^14
                    snt_fill = 1e14

                rate_info[self.scattering_labels[s]].append(snt_fill)
                self.scattering_rates[spin][s, n, t, mask] = snt_fill

            if len(self.spins) == 1:
                logger.info("Filling scattering rates [s⁻¹] outside FD cutoffs with:")
            else:
                logger.info(
                    "Filling {} scattering rates [s⁻¹] outside FD cutoffs "
                    "with:".format(spin_name[spin])
                )

            headers = ["conc [cm⁻³]", "temp [K]"]
            headers += [f"{s}" for s in self.scattering_labels]
            rate_table = []
            for i, (n, t) in enumerate(np.ndindex(self.fermi_levels.shape)):
                col = [self.doping[n] * (1 / bohr_to_cm) ** 3, self.temperatures[t]]
                col += [rate_info[s][i] for s in self.scattering_labels]
                rate_table.append(col)

            table = tabulate(
                rate_table,
                headers=headers,
                numalign="right",
                stralign="center",
                floatfmt=[".2e", ".1f"] + [".2e"] * len(self.scattering_labels),
            )
            logger.info(table)

    def set_transport_properties(
        self,
        conductivity: np.ndarray,
        seebeck: np.ndarray,
        electronic_thermal_conductivity: np.ndarray,
        mobility: Optional[np.ndarray] = None,
    ):
        self.conductivity = conductivity
        self.seebeck = seebeck
        self.electronic_thermal_conductivity = electronic_thermal_conductivity
        self.mobility = mobility

    def to_dict(self, include_mesh=defaults["write_mesh"]):
        data = {
            "doping": (self.doping * cm_to_bohr ** 3).round(),
            "temperatures": self.temperatures,
            "fermi_levels": self.fermi_levels * hartree_to_ev,
            "conductivity": self.conductivity,
            "seebeck": self.seebeck,
            "electronic_thermal_conductivity": self.electronic_thermal_conductivity,
            "mobility": self.mobility,
        }

        if include_mesh:
            rates = self.scattering_rates
            energies = self.energies
            vel = self.velocities

            ir_rates = {s: r[..., self.ir_kpoints_idx] for s, r in rates.items()}
            ir_energies = {
                s: e[:, self.ir_kpoints_idx] * hartree_to_ev
                for s, e in energies.items()
            }
            ir_vel = {s: v[:, self.ir_kpoints_idx] for s, v in vel.items()}

            mesh_data = {
                "energies": ir_energies,
                "kpoints": self.kpoints,
                "ir_kpoints": self.ir_kpoints,
                "ir_to_full_kpoint_mapping": self.ir_to_full_kpoint_mapping,
                "efermi": self.intrinsic_fermi_level * hartree_to_ev,
                "vb_idx": self.vb_idx,
                "num_electrons": self.num_electrons,
                # "dos": self.dos,  # TODO: Convert dos to eV
                "velocities": ir_vel,  # TODO: convert units
                "scattering_rates": ir_rates,
                "scattering_labels": self.scattering_labels,
                "is_metal": self.is_metal,
                "fd_cutoffs": (
                    self.fd_cutoffs[0] * hartree_to_ev,
                    self.fd_cutoffs[1] * hartree_to_ev,
                ),
                "structure": get_angstrom_structure(self.structure),
                "soc": self._soc,
                "doping": data["doping"],
                "temperatures": data["temperatures"],
                "fermi_levels": data["fermi_levels"],
            }
            data["mesh"] = mesh_data
        return data

    def to_data(self):
        data = []

        triu = np.triu_indices(3)
        for n, t in np.ndindex(len(self.doping), len(self.temperatures)):
            row = [
                self.doping[n] * cm_to_bohr ** 3,
                self.temperatures[t],
                self.fermi_levels[n, t] * hartree_to_ev,
            ]
            row.extend(self.conductivity[n, t][triu])
            row.extend(self.seebeck[n, t][triu])
            row.extend(self.electronic_thermal_conductivity[n, t][triu])

            if self.mobility is not None:
                for mob in self.mobility.values():
                    row.extend(mob[n, t][triu])
            data.append(row)

        headers = ["doping[cm^-3]", "temperature[K]", "Fermi_level[eV]"]
        ds = ("xx", "xy", "xz", "yy", "yz", "zz")

        # TODO: confirm unit of kappa
        for prop, unit in [("cond", "S/m"), ("seebeck", "µV/K"), ("kappa", "?")]:
            headers.extend([f"{prop}_{d}[{unit}]" for d in ds])

        if self.mobility is not None:
            for name in self.mobility.keys():
                headers.extend([f"{name}_mobility_{d}[cm^2/V.s]" for d in ds])

        return data, headers

    def to_file(
        self,
        directory: str = ".",
        prefix: Optional[str] = None,
        write_mesh_file: bool = defaults["write_mesh"],
        file_format: str = defaults["file_format"],
        suffix_mesh: bool = True,
    ):
        if self.conductivity is None:
            raise ValueError("Can't write AmsetData, transport properties not set")

        if not prefix:
            prefix = ""
        else:
            prefix += "_"

        if suffix_mesh:
            suffix = "_{}".format("x".join(map(str, self.kpoint_mesh)))
        else:
            suffix = ""

        if file_format in ["json", "yaml"]:
            data = self.to_dict()
            data = cast_dict_list(data)

            filename = joinpath(directory, f"{prefix}transport{suffix}.{file_format}")
            dumpfn(data, filename, indent=4)

        elif file_format in ["csv", "txt"]:
            # don't write the data as JSON, instead write raw text files
            data, headers = self.to_data()
            filename = joinpath(directory, f"{prefix}transport{suffix}.{file_format}")
            np.savetxt(filename, data, header=" ".join(headers))
        else:
            raise ValueError(f"Unrecognised output format: {file_format}")

        if write_mesh_file:
            mesh_data = self.to_dict(include_mesh=True)["mesh"]
            mesh_filename = joinpath(directory, f"{prefix}mesh{suffix}.h5")
            write_mesh(mesh_data, filename=mesh_filename)
            return filename, mesh_filename
        else:
            return filename


def check_nbands_equal(interpolator, amset_data):
    nbands_equal = [
        amset_data.energies[s].shape[0] == interpolator.nbands[s]
        for s in amset_data.spins
    ]
    return np.all(nbands_equal)



import sys
from pathlib import Path

import click
import numpy as np

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"


@click.command()
@click.option("-v", "--vasprun", default="vasprun.xml", help="vasprun.xml file")
@click.option("-o", "--outcar", default="OUTCAR", help="OUTCAR file")
def phonon_frequency(vasprun, outcar):
    """Extract the effective phonon frequency from a VASP calculation"""
    from pymatgen.io.vasp import Outcar, Vasprun
    from tabulate import tabulate

    vasprun = get_file(vasprun, Vasprun)
    outcar = get_file(outcar, Outcar)

    effective_frequency, weights, freqs = effective_phonon_frequency_from_vasp_files(
        vasprun, outcar
    )

    table = tabulate(
        list(zip(freqs, weights)),
        headers=("Frequency", "Weight"),
        numalign="right",
        stralign="center",
        floatfmt=(".2f", ".2f"),
    )
    click.echo(table)
    click.echo(f"\npop_frequency: {effective_frequency:.2f} THz")

    return effective_frequency


def effective_phonon_frequency_from_vasp_files(vasprun, outcar):
    eigenvalues = -vasprun.normalmode_eigenvals[::-1]
    eigenvectors = vasprun.normalmode_eigenvecs[::-1]

    outcar.read_lepsilon()
    born_effective_charges = outcar.born

    effective_frequency, weights, frequencies = calculate_effective_phonon_frequency(
        eigenvalues, eigenvectors, born_effective_charges, vasprun.final_structure
    )

    return effective_frequency, weights, frequencies


def calculate_effective_phonon_frequency(
    eigenvalues: np.ndarray,
    eigenvectors: np.ndarray,
    born_effecitve_charges: np.ndarray,
    structure,
):
    # get frequencies from eigenvals and convert to THz
    frequencies = 15.633302 * np.sqrt(np.abs(eigenvalues)) * np.sign(eigenvalues)

    weights = []
    for eigenvector, eigenvalue in zip(eigenvectors, eigenvalues):
        weight = get_phonon_weight(
            eigenvector, eigenvalue, born_effecitve_charges, structure
        )
        weights.append(weight)

    weights = np.array(weights) / sum(weights)
    effective_frequency = np.sum(weights * frequencies)
    return effective_frequency, weights, frequencies


def get_phonon_weight(eigenvector, eigenvalue, born_effective_charges, structure):
    from pymatgen.core.tensors import DEFAULT_QUAD

    # take spherical average of weight on scaled unit sphere
    directions = DEFAULT_QUAD["points"] * 0.01
    quad_weights = DEFAULT_QUAD["weights"]

    # precalculate Z*.e(q) / sqrt(M*omega) for each atom; this only needs to be
    # computed once and is the same for every q direction
    zes = []
    for atom_born, atom_vec, site in zip(
        born_effective_charges, eigenvector, structure
    ):
        ze = np.dot(atom_born, atom_vec) / np.sqrt(site.specie.atomic_mass * eigenvalue)
        zes.append(ze)

    all_weights = []
    for direction in directions:
        # Calculate q.Z*.e(q) / sqrt(M*omega) for each atom
        qze = np.dot(zes, direction)

        # sum across all atoms
        all_weights.append(np.abs(np.sum(qze)))

    weight = np.average(all_weights * quad_weights)

    if np.isnan(weight):
        return 0
    else:
        return weight


def get_file(filename, class_type):
    if isinstance(filename, str):
        filename_gz = filename + ".gz"

        if Path(filename).exists():
            return class_type(filename)

        elif Path(filename_gz).exists():
            return class_type(filename_gz)

        else:
            print(f"Could not find {filename}. Try running with -h option")
            sys.exit()

    elif isinstance(filename, class_type):
        return filename



import warnings

from amset.core.run import Runner
from amset.plot.rates import RatesPlotter

warnings.simplefilter("ignore")

settings = {
    # general settings
    "doping": [-3e13],
    "temperatures": [201, 290, 401, 506, 605, 789, 994],
    "bandgap": 1.33,
    # electronic_structure settings
    "interpolation_factor": 50,
    # scattering rate settings
    "deformation_potential": (1.2, 8.6),
    "elastic_constant": 139.7,
    "donor_charge": 1,
    "acceptor_charge": 1,
    "static_dielectric": 12.18,
    "high_frequency_dielectric": 10.32,
    "pop_frequency": 8.16,
    # performance settings
    "mobility_rates_only": True,
    "write_mesh": True,
}


if __name__ == "__main__":
    runner = Runner.from_vasprun("vasprun.xml.gz", settings)
    amset_data = runner.run()

    plotter = RatesPlotter(amset_data)
    plt = plotter.get_plot()
    plt.savefig("GaAs_rates.png", bbox_inches="tight", dpi=400)



import colorsys
from copy import deepcopy

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import to_rgb

from amset.constants import boltzmann_ev
from amset.electronic_structure.fd import dfdde
from amset.plot import get_figsize, styled_plot
from amset.plot.base import BaseMeshPlotter, amset_base_style, base_total_color
from amset.plot.transport import fancy_format_doping, fancy_format_temp, get_lim

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"


_legend_kwargs = {"loc": "upper left", "bbox_to_anchor": (1, 1), "frameon": False}
_fmt_data = {
    "rate": {
        "label": "Scattering rate",
        "label_symbol": r"$\tau^{-1}$",
        "unit": "s$^{-1}$",
    },
    "lifetime": {
        "label": "Scattering lifetime",
        "label_symbol": r"$\tau$",
        "unit": "s",
    },
    "v2tau": {"label": r"$v^2\tau$", "label_symbol": r"$v^2\tau$", "unit": r"cm$^2$/s"},
    "v2taudfde": {
        "label": r"$v^2\tau \left[ - \frac{\mathrm{d}f}{\mathrm{d}\varepsilon} \right ]$",
        "label_symbol": r"$v^2\tau \left[ - \frac{\mathrm{d}f}{\mathrm{d}\varepsilon} \right ]$",
        "unit": r"cm$^2$/s",
    },
    "energy": {"label": "Energy (eV)", "label_symbol": r"$\varepsilon$ (eV)"},
}


class RatesPlotter(BaseMeshPlotter):
    def __init__(self, *args, pad=0.1, use_symbol=False, **kwargs):
        super().__init__(*args, **kwargs)
        self.plot_energies = np.vstack([self.energies[s] for s in self.spins])
        self.plot_velocities = np.vstack([self.velocities[s] for s in self.spins])
        self.plot_norm_velocities = np.linalg.norm(self.plot_velocities, axis=-1)
        self.plot_rates = np.concatenate(
            [self.scattering_rates[s] for s in self.spins], axis=3
        )
        self.label_key = "label_symbol" if use_symbol else "label"
        self.pad = pad

    @styled_plot(amset_base_style)
    def get_plot(
        self,
        plot_fd_tols: bool = False,
        plot_total_rate: bool = False,
        ymin: float = None,
        ymax: float = None,
        xmin: float = None,
        xmax: float = None,
        normalize_energy: bool = True,
        separate_rates: bool = True,
        doping_idx=0,
        temperature_idx=0,
        legend=True,
        total_color=None,
        show_dfde=False,
        plot_type="rate",
        title=True,
        axes=None,
        style=None,
        no_base_style=False,
        fonts=None,
    ):
        if normalize_energy and self.is_metal:
            norm_e = self.fermi_levels[0][0]
        elif normalize_energy:
            cb_idx = {s: v + 1 for s, v in self.vb_idx.items()}
            norm_e = np.max([self.energies[s][: cb_idx[s]] for s in self.spins])
        else:
            norm_e = 0

        if doping_idx is None and len(self.doping) == 1:
            doping_idx = 0

        if temperature_idx is None and len(self.temperatures) == 1:
            temperature_idx = 0

        if doping_idx is None and temperature_idx is None:
            ny = len(self.doping)
            nx = len(self.temperatures)
            p_idx = [[(x, y) for y in range(nx)] for x in range(ny)]
        elif doping_idx is None:
            nx = len(self.doping)
            ny = 1
            p_idx = [[(x, temperature_idx) for x in range(nx)]]
        elif temperature_idx is None:
            nx = len(self.temperatures)
            ny = 1
            p_idx = [[(doping_idx, x) for x in range(nx)]]
        else:
            nx = 1
            ny = 1
            p_idx = [[(doping_idx, temperature_idx)]]

        if axes is None:
            _, axes = plt.subplots(ny, nx, figsize=get_figsize(ny, nx))

        for y_idx, p_x_idx in enumerate(p_idx):
            for x_idx, (n, t) in enumerate(p_x_idx):
                if nx == 1 and ny == 1:
                    ax = axes
                elif ny == 1:
                    ax = axes[x_idx]
                else:
                    ax = axes[y_idx, x_idx]

                if x_idx == nx - 1 and y_idx == ny - 1 and legend:
                    show_legend = True
                else:
                    show_legend = False

                doping_fmt = fancy_format_doping(self.doping[n])
                temp_fmt = fancy_format_temp(self.temperatures[t])
                title_str = f"{doping_fmt}    {temp_fmt}"

                self.plot_rates_to_axis(
                    ax,
                    n,
                    t,
                    plot_type=plot_type,
                    separate_rates=separate_rates,
                    plot_total_rate=plot_total_rate,
                    plot_fd_tols=plot_fd_tols,
                    ymin=ymin,
                    ymax=ymax,
                    xmin=xmin,
                    xmax=xmax,
                    show_legend=show_legend,
                    legend_kwargs=_legend_kwargs,
                    normalize_energy=norm_e,
                    total_color=total_color,
                    show_dfde=show_dfde,
                )
                if title:
                    ax.set(title=title_str)

        return plt

    def plot_rates_to_axis(
        self,
        ax,
        doping_idx,
        temperature_idx,
        plot_type="rate",
        separate_rates=True,
        plot_total_rate: bool = False,
        plot_fd_tols: bool = True,
        show_legend: bool = True,
        legend_kwargs=None,
        ymin=None,
        ymax=None,
        xmin=None,
        xmax=None,
        normalize_energy=0,
        scaling_factor=1,
        total_color=None,
        show_dfde=False,
    ):
        rates = self.plot_rates[:, doping_idx, temperature_idx]
        if separate_rates:
            sort_idx = np.argsort(self.scattering_labels)
            labels = self.scattering_labels[sort_idx].tolist()
            rates = rates[sort_idx]
        else:
            rates = np.sum(rates, axis=0)[None, ...]
            labels = ["rates"]

        if legend_kwargs is None:
            legend_kwargs = {}

        labels = deepcopy(labels)
        if plot_total_rate:
            # add total rates column
            rates = np.concatenate((rates, np.sum(rates, axis=0)[None, ...]))
            labels += ["total"]

        min_fd = self.fd_cutoffs[0]
        max_fd = self.fd_cutoffs[1]
        if not xmin:
            min_e = min_fd
        else:
            min_e = xmin + normalize_energy

        if not xmax:
            max_e = max_fd
        else:
            max_e = xmax + normalize_energy

        energies = self.plot_energies.ravel()
        energy_mask = (energies > min_e) & (energies < max_e)
        energies = energies[energy_mask]

        dfde_occ = -dfdde(
            energies,
            self.fermi_levels[doping_idx, temperature_idx],
            boltzmann_ev * self.temperatures[temperature_idx],
        )

        plt_rates = {}
        for label, rate in zip(labels, rates):
            rate = rate.ravel()[energy_mask]
            if plot_type == "lifetime":
                rate = 1 / rate
            elif plot_type == "v2tau":
                rate = self.plot_norm_velocities.ravel()[energy_mask] ** 2 / rate
            elif plot_type == "v2taudfde":
                rate = (
                    self.plot_norm_velocities.ravel()[energy_mask] ** 2
                    * dfde_occ
                    / rate
                )
            elif plot_type == "rate":
                pass
            else:
                raise ValueError(f"Unknown plot_type: {plot_type}")

            rate *= scaling_factor
            plt_rates[label] = rate

        rates_in_cutoffs = np.array(list(plt_rates.values()))
        ylim = get_lim(
            rates_in_cutoffs[rates_in_cutoffs > 0], ymin, ymax, True, self.pad
        )

        # convert energies to eV and normalise
        norm_energies = energies - normalize_energy
        norm_min_fd = min_fd - normalize_energy
        norm_max_fd = max_fd - normalize_energy
        xlim = (min_e - normalize_energy, max_e - normalize_energy)

        if total_color is None:
            total_color = base_total_color

        colors = matplotlib.rcParams["axes.prop_cycle"].by_key()["color"]
        for i, (label, rate) in enumerate(plt_rates.items()):
            if label == "total":
                c = total_color
            else:
                c = colors[i]

            legend_c = c
            if show_dfde:
                c = _get_lightened_colors(c, dfde_occ / np.max(dfde_occ))
                # sort on lightness to put darker points on top of lighter ones
                sort_idx = np.argsort(dfde_occ)
                norm_energies_sort = norm_energies[sort_idx]
                rate = rate[sort_idx]
                c = c[sort_idx]
            else:
                norm_energies_sort = norm_energies

            ax.scatter(norm_energies_sort, rate, c=c, rasterized=True)
            # hack to plot get the correct label if using dfde occupation
            ax.scatter(-10000000, 0, c=legend_c, label=label)

        if plot_fd_tols:
            ax.plot((norm_min_fd, norm_min_fd), ylim, c="gray", ls="--", alpha=0.5)
            ax.plot(
                (norm_max_fd, norm_max_fd),
                ylim,
                c="gray",
                ls="--",
                alpha=0.5,
                label="FD cutoffs",
            )

        scaling_string = f"{scaling_factor:g} " if scaling_factor != 1 else ""
        data_label = _fmt_data[plot_type][self.label_key]
        data_unit = _fmt_data[plot_type]["unit"]
        ylabel = f"{data_label} ({scaling_string}{data_unit})"
        xlabel = _fmt_data["energy"][self.label_key]

        ax.set(xlabel=xlabel, ylabel=ylabel, ylim=ylim, xlim=xlim)
        ax.semilogy()

        if show_legend:
            ax.legend(**legend_kwargs)


def _get_lightened_colors(color, occupation, min_saturation=0, min_lightness=0.92):
    color_rgb = np.array(to_rgb(color))
    h, ll, s = colorsys.rgb_to_hls(*color_rgb)

    lighten_amount = 1 - occupation
    s = s - (s - min_saturation) * lighten_amount
    ll = ll - (ll - min_lightness) * lighten_amount

    light_colors_rgb = []
    for s_new, l_new in zip(s, ll):
        light_colors_rgb.append(colorsys.hls_to_rgb(h, l_new, s_new))
    return np.array(light_colors_rgb)



import logging
from pathlib import Path

from monty.serialization import loadfn
from pymatgen.electronic_structure.bandstructure import BandStructure
from sumo.plotting.bs_plotter import SBSPlotter
from sumo.plotting.dos_plotter import SDOSPlotter

from amset.constants import defaults
from amset.electronic_structure.common import get_band_structure
from amset.interpolation.bandstructure import Interpolator
from amset.log import initialize_amset_logger

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"

logger = logging.getLogger(__name__)


class ElectronicStructurePlotter:
    def __init__(
        self,
        bandstructure: BandStructure,
        nelect: int,
        soc: bool = False,
        interpolation_factor=defaults["interpolation_factor"],
        print_log=defaults["print_log"],
        symprec=defaults["symprec"],
        energy_cutoff=None,
    ):
        if print_log:
            initialize_amset_logger(filename=False)

        self.symprec = symprec
        self.interpolater = Interpolator(
            bandstructure, nelect, interpolation_factor=interpolation_factor, soc=soc
        )
        self.structure = bandstructure.structure
        self.energy_cutoff = energy_cutoff

    @classmethod
    def from_band_structure_data(cls, filename, **kwargs):
        data = loadfn(filename)
        soc = data.get("soc", False)
        return cls(data["band_structure"], data["nelect"], soc=soc, **kwargs)

    @classmethod
    def from_vasprun(
        cls, vasprun, zero_weighted_kpoints=defaults["zero_weighted_kpoints"], **kwargs
    ):
        from pymatgen.io.vasp import Vasprun

        if isinstance(vasprun, Path):
            vasprun = vasprun.as_posix()

        if isinstance(vasprun, str):
            vasprun_gz = vasprun + ".gz"

            if Path(vasprun).exists():
                vasprun = Vasprun(vasprun)

            elif Path(vasprun_gz).exists():
                vasprun = Vasprun(vasprun_gz)

        bandstructure = get_band_structure(vasprun, zero_weighted=zero_weighted_kpoints)
        nelect = vasprun.parameters["NELECT"]
        soc = vasprun.parameters["LSORBIT"]
        return cls(bandstructure, nelect, soc=soc, **kwargs)

    def get_plot(
        self,
        plot_dos=True,
        plot_band_structure=True,
        kpath=None,
        line_density=100.0,
        dos_kpoints=50.0,
        dos_estep=defaults["dos_estep"],
        zero_to_efermi=True,
        vbm_cbm_marker=False,
        height=6,
        width=None,
        emin=None,
        emax=None,
        elabel="Energy (eV)",
        plt=None,
        dos_aspect=3,
        return_plotter=False,
        aspect=None,
        style=None,
        no_base_style=False,
        fonts=None,
    ):
        if width is None and plot_band_structure:
            width = 6

        if self.energy_cutoff is None:
            self.energy_cutoff = max([abs(emin or 6), abs(emax or 6)])

        if plot_band_structure:
            logger.info("Generating band structure...")
            bs_plotter = self.get_bs_plotter(line_density=line_density, kpath=kpath)

        if plot_dos:
            logger.info("Generating density of states...")
            dos_plotter = self.get_dos_plotter(
                dos_density=dos_kpoints, dos_estep=dos_estep
            )

        logger.info("Plotting...")
        common_kwargs = dict(
            zero_to_efermi=zero_to_efermi,
            width=width,
            height=height,
            plt=plt,
            style=style,
            no_base_style=no_base_style,
            fonts=fonts,
        )
        if plot_band_structure:
            bs_kwargs = {
                "ymin": emin,
                "ymax": emax,
                "ylabel": elabel,
                "aspect": aspect,
                "vbm_cbm_marker": vbm_cbm_marker,
            }
            common_kwargs.update(bs_kwargs)

            if plot_dos:
                bs_dos_kwargs = {
                    # "dos_label": "DOS",
                    "dos_aspect": dos_aspect,
                    "dos_plotter": dos_plotter,
                    "plot_dos_legend": False,
                }
                common_kwargs.update(bs_dos_kwargs)

            if return_plotter:
                return bs_plotter.get_plot(**common_kwargs), bs_plotter
            else:
                return bs_plotter.get_plot(**common_kwargs)

        else:
            dos_kwargs = {
                "xmin": emin,
                "xmax": emax,
                "legend_on": False,
                "xlabel": elabel,
                "ylabel": "DOS",
            }
            common_kwargs.update(dos_kwargs)

            if return_plotter:
                return dos_plotter.get_plot(common_kwargs), dos_plotter
            else:
                return dos_plotter.get_plot(common_kwargs)

    def get_bs_plotter(self, line_density=100, kpath=None):
        lm_bs = self.interpolater.get_line_mode_band_structure(
            line_density=line_density,
            kpath=kpath,
            symprec=self.symprec,
            energy_cutoff=self.energy_cutoff,
        )
        return SBSPlotter(lm_bs)

    def get_dos_plotter(self, dos_density=50.0, dos_estep=defaults["dos_estep"]):
        dos = self.interpolater.get_dos(
            kpoint_mesh=dos_density,
            symprec=self.symprec,
            estep=dos_estep,
            energy_cutoff=self.energy_cutoff,
        )

        return SDOSPlotter(dos, {})



import logging

import numba
import numpy as np
from interpolation.splines import eval_linear
from interpolation.splines import extrap_options as xto

from amset.constants import defaults
from amset.electronic_structure.kpoints import get_mesh_from_kpoint_numbers
from amset.electronic_structure.symmetry import expand_kpoints
from amset.interpolation.periodic import (
    PeriodicLinearInterpolator,
    group_bands_and_kpoints,
)
from amset.util import array_from_buffer, create_shared_array
from amset.wavefunction.common import desymmetrize_coefficients, is_ncl
from amset.wavefunction.io import load_coefficients

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"

logger = logging.getLogger(__name__)


class WavefunctionOverlapCalculator(PeriodicLinearInterpolator):
    def __init__(self, nbands, data_shape, interpolators, ncl, gpoints):
        super().__init__(nbands, data_shape, interpolators)
        self.ncl = ncl
        self.gpoints = gpoints

    def to_reference(self):
        interpolator_references = self._interpolators_to_reference()
        gpoints_buffer, self.gpoints = create_shared_array(
            self.gpoints, return_shared_data=True
        )
        return (
            self.nbands,
            self.data_shape,
            interpolator_references,
            self.ncl,
            gpoints_buffer,
        )

    @classmethod
    def from_reference(
        cls, nbands, data_shape, interpolator_references, ncl, gpoints_buffer
    ):
        interpolators = cls._interpolators_from_reference(interpolator_references)
        gpoints = array_from_buffer(gpoints_buffer)
        return cls(nbands, data_shape, interpolators, ncl, gpoints)

    @classmethod
    def from_data(cls, kpoints, data, gpoints=None, gaussian=None):
        if gpoints is None:
            raise ValueError("gpoints required for initialization")
        ncl = is_ncl(data)
        grid_kpoints, mesh_dim, sort_idx = cls._grid_kpoints(kpoints)
        nbands, data_shape, interpolators = cls._setup_interpolators(
            data, grid_kpoints, mesh_dim, sort_idx, gaussian
        )
        return cls(nbands, data_shape, interpolators, ncl, gpoints)

    @classmethod
    def from_file(cls, filename):
        coeff, gpoints, kpoints, structure = load_coefficients(filename)
        return cls.from_coefficients(coeff, gpoints, kpoints, structure)

    @classmethod
    def from_coefficients(
        cls, coefficients, gpoints, kpoints, structure, symprec=defaults["symprec"]
    ):
        logger.info("Initializing wavefunction overlap calculator")

        mesh_dim = get_mesh_from_kpoint_numbers(kpoints)
        if np.product(mesh_dim) == len(kpoints):
            return cls.from_data(kpoints, coefficients, gpoints)

        full_kpoints, *symmetry_mapping = expand_kpoints(
            structure, kpoints, time_reversal=True, return_mapping=True, symprec=symprec
        )
        coefficients = desymmetrize_coefficients(
            coefficients, gpoints, kpoints, structure, *symmetry_mapping
        )
        return cls.from_data(full_kpoints, coefficients, gpoints)

    def get_coefficients(self, spin, bands, kpoints):
        interp_coeffs = self.interpolate(spin, bands, kpoints)
        if self.ncl:
            interp_coeffs /= np.linalg.norm(interp_coeffs, axis=(-2, -1))[
                ..., None, None
            ]
        else:
            interp_coeffs /= np.linalg.norm(interp_coeffs, axis=-1)[:, None]
        return interp_coeffs

    def get_overlap(self, spin, band_a, kpoint_a, band_b, kpoint_b):
        # generally, we don't want to do the interpolation for all band and
        # k-points simultaneously as this can use a lot of memory. This becomes
        # an issue when using multiprocessing. I.e., if you're parralellising over
        # 24 cores on a single node, you only have access 1/24 the total memory
        # for each core. Here we use numba jit to calculate the overlap of each
        # band/kpoint sequentially. In testing this only gives ~2x slow down vs
        # the simultaneous approach, while using an order of magnitude less memory.
        bands, kpoints, single_overlap = group_bands_and_kpoints(
            band_a, kpoint_a, band_b, kpoint_b
        )
        grid, data = self.interpolators[spin]
        v = np.concatenate([np.asarray(bands)[:, None], np.asarray(kpoints)], axis=1)

        if self.ncl:
            overlap = _get_overlap_ncl(grid, data, v, self.data_shape[0])
        else:
            overlap = _get_overlap(grid, data, v, self.data_shape[0])

        if single_overlap:
            return overlap[0]
        else:
            return overlap


@numba.njit
def _get_overlap(grid, data, points, n_coeffs):
    initial = np.zeros(n_coeffs, dtype=np.complex64)
    initial.real[:] = eval_linear(grid, data.real, points[0], xto.LINEAR)
    initial.imag[:] = eval_linear(grid, data.imag, points[0], xto.LINEAR)
    initial /= np.linalg.norm(initial)
    initial[:] = np.conj(initial)

    res = np.zeros(points.shape[0] - 1)
    final = np.zeros(n_coeffs, dtype=np.complex64)
    for i in range(1, points.shape[0]):
        final.real[:] = eval_linear(grid, data.real, points[i], xto.LINEAR)
        final.imag[:] = eval_linear(grid, data.imag, points[i], xto.LINEAR)
        final /= np.linalg.norm(final)
        res[i - 1] = np.abs(np.dot(final, initial)) ** 2
    return res


@numba.njit
def _get_overlap_ncl(grid, data, points, n_coeffs):
    initial = np.zeros((n_coeffs, 2), dtype=np.complex64)
    initial.real[:] = eval_linear(grid, data.real, points[0], xto.LINEAR).reshape(
        (n_coeffs, 2)
    )
    initial.imag[:] = eval_linear(grid, data.imag, points[0], xto.LINEAR).reshape(
        (n_coeffs, 2)
    )
    initial /= np.linalg.norm(initial)
    initial[:] = np.conj(initial)

    res = np.zeros(points.shape[0] - 1)
    final = np.zeros((n_coeffs, 2), dtype=np.complex64)
    for i in range(1, points.shape[0]):
        final.real[:] = eval_linear(grid, data.real, points[i], xto.LINEAR).reshape(
            (n_coeffs, 2)
        )
        final.imag[:] = eval_linear(grid, data.imag, points[i], xto.LINEAR).reshape(
            (n_coeffs, 2)
        )
        final /= np.linalg.norm(final)

        sum_ = 0j
        for j in range(final.shape[0]):
            sum_ += initial[j, 0] * final[j, 0] + initial[j, 1] * final[j, 1]

        res[i - 1] = abs(sum_) ** 2

    return res



import pytest
from pymatgen.electronic_structure.core import Spin

from amset.electronic_structure.common import (
    get_cbm_energy,
    get_efermi,
    get_vb_idx,
    get_vbm_energy,
)


@pytest.mark.parametrize(
    "system,energy_cutoff,expected",
    [
        pytest.param(
            "tricky_sp", 1.5, {Spin.up: 2, Spin.down: -1}, id="default tricky"
        ),
        pytest.param("tricky_sp", 5, {Spin.up: 12, Spin.down: 10}, id="large tricky"),
    ],
)
def test_get_vb_idx_tricky(band_structures, system, energy_cutoff, expected):
    result = get_vb_idx(energy_cutoff, band_structures[system])
    assert result == expected


@pytest.mark.parametrize(
    "system,vb_idx,expected",
    [
        pytest.param(
            "tricky_sp", {Spin.up: 37, Spin.down: 34}, 0.6084, id="both spin tricky"
        ),
        pytest.param(
            "tricky_sp", {Spin.up: 37, Spin.down: -1}, 0.6084, id="up only tricky"
        ),
        pytest.param(
            "tricky_sp", {Spin.up: -1, Spin.down: 34}, -2.9677, id="down only tricky"
        ),
    ],
)
def test_get_vbm_energy(band_structures, system, vb_idx, expected):
    result = get_vbm_energy(band_structures[system].bands, vb_idx)
    assert result == expected


@pytest.mark.parametrize(
    "system,vb_idx,expected",
    [
        pytest.param(
            "tricky_sp", {Spin.up: 38, Spin.down: 35}, 2.3641, id="both spin tricky"
        ),
        pytest.param(
            "tricky_sp", {Spin.up: 38, Spin.down: 79}, 4.434, id="up only tricky"
        ),
        pytest.param(
            "tricky_sp", {Spin.up: 79, Spin.down: 35}, 2.3641, id="down only tricky"
        ),
    ],
)
def test_get_cbm_energy(band_structures, system, vb_idx, expected):
    result = get_cbm_energy(band_structures[system].bands, vb_idx)
    assert result == expected


@pytest.mark.parametrize(
    "system,vb_idx,expected",
    [
        pytest.param(
            "tricky_sp", {Spin.up: 38, Spin.down: 35}, 3.4574, id="both spin tricky"
        ),
        pytest.param(
            "tricky_sp", {Spin.up: 38, Spin.down: 79}, 14.50355, id="up only tricky"
        ),
        pytest.param(
            "tricky_sp", {Spin.up: 79, Spin.down: 35}, 13.4251, id="down only tricky"
        ),
    ],
)
def test_get_efermi(band_structures, system, vb_idx, expected):
    result = get_efermi(band_structures[system].bands, vb_idx)
    assert result == expected



from pathlib import Path

from amset.constants import defaults
from amset.io import load_settings, write_settings


def test_load_settings(test_dir):
    settings = load_settings(test_dir / "amset_settings.yaml")

    # test settings loaded correctly
    assert settings["scissor"] == 3.0
    assert settings["high_frequency_dielectric"].tolist() == [
        [10, 0, 0],
        [0, 10, 0],
        [0, 0, 10],
    ]

    # test defaults inferred correctly
    assert settings["pop_frequency"] == defaults["pop_frequency"]
    assert settings["dos_estep"] == defaults["dos_estep"]


def test_write_settings(clean_dir):
    settings_file = Path("test_settings.yaml")
    write_settings(defaults, settings_file)
    contents = settings_file.read_text()

    assert "scattering_type: auto" in contents
    assert "defect_charge: 1" in contents



from pathlib import Path

import numpy as np

from amset.plot.base import PlotData, write_plot_data


def test_plot_data(clean_dir):
    # test initialization
    plot_data = PlotData(
        x_property="doping",
        y_property="cond",
        x_label="Carrier concentration [cm$^{3}$]",
        y_label="Conductivity [S/m]",
        labels=["T = 100 K", "T = 200 K"],
        x=np.array([1e16, 1e17, 1e18]),
        y=np.array([[1, 2, 3], [4, 5, 6]]),
    )

    # test comment
    assert "x-axis: Carrier concentration" in plot_data.comment
    assert "y-axis: Conductivity" in plot_data.comment
    assert "T=100K" in plot_data.comment

    # test data
    assert plot_data.data.shape == (3, 3)
    assert plot_data.data[0, 1] == 1e17
    assert plot_data.data[2, 1] == 5

    # test writing to file
    filename = Path("test_data.dat")
    plot_data.to_file(filename)
    assert filename.exists()
    assert "x-axis: Carrier concentration" in filename.read_text()


def test_write_plot_data(clean_dir):
    # test initialization
    plot_data = [
        PlotData(
            x_property="doping",
            y_property="cond",
            x_label="Carrier concentration [cm$^{3}$]",
            y_label="Conductivity [S/m]",
            labels=["T = 100 K", "T = 200 K"],
            x=np.array([1e16, 1e17, 1e18]),
            y=np.array([[1, 2, 3], [4, 5, 99]]),
        ),
        PlotData(
            x_property="doping",
            y_property="seebeck",
            x_label="Carrier concentration [cm$^{3}$]",
            y_label="Seebeck coefficient [S/m]",
            labels=["T = 100 K", "T = 200 K"],
            x=np.array([1e16, 1e17, 1e18]),
            y=np.array([[10, 11, 12], [15, 16, 101]]),
        ),
    ]

    files = [
        {
            "filename": "cond.dat",
            "should_contain": ["y-axis: Conductivity", " 9.90000000e+01"],
        },
        {
            "filename": "seebeck.dat",
            "should_contain": ["y-axis: Seebeck coefficient", " 1.01000000e+02"],
        },
    ]

    # basic write
    write_plot_data(plot_data)
    for file_data in files:
        filename = Path(file_data["filename"])
        assert filename.exists(), f"{filename} does not exist"

        contents = filename.read_text()
        for expected in file_data["should_contain"]:
            assert expected in contents, f"{filename} does not contain {expected}"

    # write with prefix
    prefix = "test_"
    write_plot_data(plot_data, prefix=prefix)
    for file_data in files:
        filename = Path(prefix + file_data["filename"])
        assert filename.exists(), f"{filename} does not exist"

        contents = filename.read_text()
        for expected in file_data["should_contain"]:
            assert expected in contents, f"{filename} does not contain {expected}"



from contextlib import contextmanager

import numpy as np
import pytest
from pymatgen.electronic_structure.core import Spin

from amset.util import (
    cast_dict_list,
    cast_dict_ndarray,
    cast_elastic_tensor,
    cast_piezoelectric_tensor,
    cast_tensor,
    get_progress_bar,
    groupby,
    parse_deformation_potential,
    parse_doping,
    parse_ibands,
    parse_temperatures,
    tensor_average,
    validate_settings,
)


@contextmanager
def does_not_raise():
    yield


@pytest.mark.parametrize(
    "tensor,expected",
    [
        pytest.param([[3, 0, 0], [0, 4, 0], [0, 0, 5]], 4, id="diagonal"),
        pytest.param([[0, 3, 3], [3, 3, 4], [3, 4, 3]], 2, id="off-diagonal"),
    ],
)
def test_tensor_average(tensor, expected):
    assert tensor_average(tensor) == expected


@pytest.mark.parametrize(
    "elements, groups, expected",
    [
        pytest.param(
            ["a", "b", "1", "2", "c", "d"],
            [2, 0, 1, 2, 0, 0],
            [["b", "c", "d"], ["1"], ["a", "2"]],
            id="mixed",
        ),
        pytest.param(
            [[0, 0, 0], [1, 1, 1], [2, 2, 2], [3, 3, 3], [4, 4, 4]],
            [2, 2, 1, 0, 0],
            [[[3, 3, 3], [4, 4, 4]], [[2, 2, 2]], [[0, 0, 0], [1, 1, 1]]],
            id="coords",
        ),
    ],
)
def test_groupby(elements, groups, expected):
    elements = ["a", "b", "1", "2", "c", "d"]
    groups = [2, 0, 1, 2, 0, 0]
    expected_output = [["b", "c", "d"], ["1"], ["a", "2"]]
    output = groupby(elements, groups)
    output = [x.tolist() for x in output]
    assert output == expected_output


_expected_elastic = [
    [
        [[3.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]],
        [[0.0, 1.5, 0.0], [1.5, 0.0, 0.0], [0.0, 0.0, 0.0]],
        [[0.0, 0.0, 1.5], [0.0, 0.0, 0.0], [1.5, 0.0, 0.0]],
    ],
    [
        [[0.0, 1.5, 0.0], [1.5, 0.0, 0.0], [0.0, 0.0, 0.0]],
        [[0.0, 0.0, 0.0], [0.0, 3.0, 0.0], [0.0, 0.0, 0.0]],
        [[0.0, 0.0, 0.0], [0.0, 0.0, 1.5], [0.0, 1.5, 0.0]],
    ],
    [
        [[0.0, 0.0, 1.5], [0.0, 0.0, 0.0], [1.5, 0.0, 0.0]],
        [[0.0, 0.0, 0.0], [0.0, 0.0, 1.5], [0.0, 1.5, 0.0]],
        [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 3.0]],
    ],
]
_elastic_voigt = [
    [3, 0, 0, 0, 0, 0],
    [0, 3, 0, 0, 0, 0],
    [0, 0, 3, 0, 0, 0],
    [0, 0, 0, 1.5, 0, 0],
    [0, 0, 0, 0, 1.5, 0],
    [0, 0, 0, 0, 0, 1.5],
]
_expected_piezo = [
    [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0084], [0.0, 0.0084, 0.0]],
    [[0.0, 0.0, 0.0084], [0.0, 0.0, 0.0], [0.0084, 0.0, 0.0]],
    [[0.0, 0.0084, 0.0], [0.0084, 0.0, 0.0], [0.0, 0.0, 0.0]],
]
_piezo_voigt = [
    [0, 0, 0, 0.0084, 0, 0],
    [0, 0, 0, 0, 0.0084, 0],
    [0, 0, 0, 0, 0, 0.0084],
]


@pytest.mark.parametrize(
    "settings,expected",
    [
        pytest.param(
            {},
            {
                "scattering_type": "auto",
                "temperatures": np.array([300]),
                "calculate_mobility": True,
                "separate_mobility": True,
                "mobility_rates_only": False,
            },
            id="empty",
        ),
        pytest.param(
            {"doping": "1E16:1E20:5", "temperatures": "100:500:5"},
            {
                "doping": np.array([1e16, 1e17, 1e18, 1e19, 1e20]),
                "temperatures": np.array([100, 200, 300, 400, 500]),
            },
            id="doping",
        ),
        pytest.param(
            {"deformation_potential": "1,2"},
            {"deformation_potential": (1, 2)},
            id="deformation (str-1)",
        ),
        pytest.param(
            {"deformation_potential": "1."},
            {"deformation_potential": 1.0},
            id="deformation (str-2)",
        ),
        pytest.param(
            {"deformation_potential": "deformation.h5"},
            {"deformation_potential": "deformation.h5"},
            id="deformation (str-3)",
        ),
        pytest.param(
            {
                "static_dielectric": 3,
                "high_frequency_dielectric": 3,
                "elastic_constant": 3,
            },
            {
                "static_dielectric": np.eye(3) * 3,
                "high_frequency_dielectric": np.eye(3) * 3,
                "elastic_constant": np.array(_expected_elastic),
            },
            id="tensor cast (int)",
        ),
        pytest.param(
            {"static_dielectric": [1, 2, 3], "high_frequency_dielectric": [1, 2, 3]},
            {
                "static_dielectric": np.array([[1, 0, 0], [0, 2, 0], [0, 0, 3]]),
                "high_frequency_dielectric": np.array(
                    [[1, 0, 0], [0, 2, 0], [0, 0, 3]]
                ),
            },
            id="tensor cast (list)",
        ),
        pytest.param(
            {
                "static_dielectric": [[0, 1, 2], [3, 4, 5], [6, 7, 8]],
                "high_frequency_dielectric": [[0, 1, 2], [3, 4, 5], [6, 7, 8]],
                "elastic_constant": _expected_elastic,
                "piezoelectric_constant": _piezo_voigt,
            },
            {
                "static_dielectric": np.arange(9).reshape(3, 3),
                "high_frequency_dielectric": np.arange(9).reshape(3, 3),
                "elastic_constant": np.array(_expected_elastic),
                "piezoelectric_constant": np.array(_expected_piezo),
            },
            id="tensor cast (array)",
        ),
        pytest.param({"donor_charge": 2}, {"defect_charge": 2}, id="d charge"),
        pytest.param({"acceptor_charge": 2}, {"defect_charge": 2}, id="a charge"),
        pytest.param({"mispelt_parameter": 1}, pytest.raises(ValueError), id="raises"),
    ],
)
def test_validate_settings(settings, expected):
    if not isinstance(expected, dict):
        # ensure errors are thrown correctly
        with expected:
            validate_settings(settings)

    else:
        settings = validate_settings(settings)
        for name, expected_value in expected.items():
            value = settings[name]
            assert isinstance(settings[name], type(expected_value))

            if isinstance(value, np.ndarray):
                expected_value = expected_value.tolist()
                value = value.tolist()

            assert expected_value == value


@pytest.mark.parametrize(
    "value,expected",
    [
        pytest.param(3, np.eye(3) * 3, id="int"),
        pytest.param([0, 1, 2], [[0, 0, 0], [0, 1, 0], [0, 0, 2]], id="list"),
    ],
)
def test_cast_tensor(value, expected):
    np.testing.assert_array_equal(cast_tensor(value), expected)


@pytest.mark.parametrize(
    "value,expected",
    [
        pytest.param(3, _expected_elastic, id="int"),
        pytest.param(_elastic_voigt, _expected_elastic, id="Voigt"),
        pytest.param(_expected_elastic, _expected_elastic, id="3x3x3x3"),
    ],
)
def test_cast_elastic_tensor(value, expected):
    np.testing.assert_array_equal(cast_elastic_tensor(value), expected)


@pytest.mark.parametrize(
    "value,expected",
    [
        pytest.param(_piezo_voigt, _expected_piezo, id="Voigt"),
        pytest.param(_expected_piezo, _expected_piezo, id="3x3x3"),
    ],
)
def test_cast_piezoelectric_tensor(value, expected):
    np.testing.assert_array_equal(cast_piezoelectric_tensor(value), expected)


@pytest.mark.parametrize(
    "value,expected",
    [
        pytest.param(
            {"a": np.array([1, 2, 3]), "c": [1]},
            {"a": [1, 2, 3], "c": [1]},
            id="single",
        ),
        pytest.param(
            {"a": np.array([1, 2, 3]), "c": [1], 123: np.array([[0, 0], [0, 0]])},
            {"a": [1, 2, 3], "c": [1], 123: [[0, 0], [0, 0]]},
            id="double",
        ),
        pytest.param(
            {"a": {"b": np.array([1, 2, 3])}, "c": {"d": np.array([1, 2])}},
            {"a": {"b": [1, 2, 3]}, "c": {"d": [1, 2]}},
            id="nested",
        ),
        pytest.param(
            {"a": {Spin.up: np.array([1, 2, 3])}, "c": {"d": np.array([1, 2])}},
            {"a": {"up": [1, 2, 3]}, "c": {"d": [1, 2]}},
            id="spin",
        ),
    ],
)
def test_cast_dict_list(value, expected):
    assert cast_dict_list(value) == expected


@pytest.mark.parametrize(
    "value,expected",
    [
        pytest.param(
            {"a": [1, 2, 3], "c": 1}, {"a": np.array([1, 2, 3]), "c": 1}, id="single"
        ),
        pytest.param(
            {"a": [1, 2, 3], "c": 1, 123: [[0, 0], [0, 0]]},
            {"a": np.array([1, 2, 3]), "c": 1, 123: np.array([[0, 0], [0, 0]])},
            id="double",
        ),
        pytest.param(
            {"a": {"b": [1, 2, 3]}, "c": {"d": [1, 2]}},
            {"a": {"b": np.array([1, 2, 3])}, "c": {"d": np.array([1, 2])}},
            id="nested",
        ),
        pytest.param(
            {"a": {"up": [1, 2, 3]}, "c": {"d": [1, 2]}},
            {"a": {Spin.up: np.array([1, 2, 3])}, "c": {"d": np.array([1, 2])}},
            id="spin",
        ),
    ],
)
def test_cast_dict_ndarray(value, expected):
    def compare(a, b):
        if isinstance(a, dict) and isinstance(b, dict):
            all_keys = list(a.keys()) + list(b.keys())
            for k in all_keys:
                compare(a[k], b[k])
        elif isinstance(a, dict) or isinstance(b, dict):
            assert False
        else:
            if isinstance(a, np.ndarray) and isinstance(b, np.ndarray):
                a = a.tolist()
                b = b.tolist()
            elif isinstance(a, np.ndarray) or isinstance(b, np.ndarray):
                # both not numpy arrays
                assert False

            assert a == b

    compare(cast_dict_ndarray(value), expected)


@pytest.mark.parametrize(
    "value,expected",
    [
        pytest.param("1E16", [1e16], id="single"),
        pytest.param("1E16,1E17", [1e16, 1e17], id="multiple"),
        pytest.param("1E16:1E19:4", [1e16, 1e17, 1e18, 1e19], id="range"),
        pytest.param("1E16:1E19:4:1", pytest.raises(ValueError), id="error"),
    ],
)
def test_parse_doping(value, expected):
    if not isinstance(expected, list):
        with expected:
            parse_doping(value)
    else:
        parsed = parse_doping(value)
        assert type(parsed) == np.ndarray
        assert parsed.tolist() == expected


@pytest.mark.parametrize(
    "value,expected",
    [
        pytest.param("100", [100], id="single"),
        pytest.param("100,200", [100, 200], id="multiple"),
        pytest.param("100:400:4", [100, 200, 300, 400], id="range"),
        pytest.param("100:400:4:1", pytest.raises(ValueError), id="error"),
    ],
)
def test_parse_temperatures(value, expected):
    if not isinstance(expected, list):
        with expected:
            parse_temperatures(value)
    else:
        parsed = parse_temperatures(value)
        assert type(parsed) == np.ndarray
        assert parsed.tolist() == expected


@pytest.mark.parametrize(
    "value,expected",
    [
        pytest.param("deformation.h5", "deformation.h5", id="str"),
        pytest.param("2", 2, id="single"),
        pytest.param("2,5", (2, 5), id="both"),
        pytest.param("100,2,1", pytest.raises(ValueError), id="error"),
    ],
)
def test_parse_deformation_potential(value, expected):
    if not isinstance(expected, (tuple, int, float, str)):
        with expected:
            parse_deformation_potential(value)
    else:
        parsed = parse_deformation_potential(value)
        assert parsed == expected


@pytest.mark.parametrize(
    "iterable,total,error",
    [
        pytest.param([1, 2, 3], None, does_not_raise(), id="iterable"),
        pytest.param(None, 3, does_not_raise(), id="total"),
        pytest.param(None, None, pytest.raises(ValueError), id="error"),
    ],
)
def test_get_progress_bar(iterable, total, error):
    with error:
        pbar = get_progress_bar(iterable=iterable, total=total)
        pbar.close()


@pytest.mark.parametrize(
    "value,expected",
    [
        pytest.param("1", {Spin.up: [0]}, id="single up"),
        pytest.param("1, 2", {Spin.up: [0, 1]}, id="multiple up"),
        pytest.param("1:4", {Spin.up: [0, 1, 2, 3]}, id="range up"),
        pytest.param("1.2", {Spin.up: [0], Spin.down: [1]}, id="single up-down"),
        pytest.param(
            "1, 3, 5.2, 4, 6",
            {Spin.up: [0, 2, 4], Spin.down: [1, 3, 5]},
            id="multiple up-down",
        ),
        pytest.param(
            "1:4.2:5",
            {Spin.up: [0, 1, 2, 3], Spin.down: [1, 2, 3, 4]},
            id="range up-down",
        ),
        pytest.param([1, 2, 3, 4], {Spin.up: [0, 1, 2, 3]}, id="list up"),
        pytest.param(
            ([1, 2, 3, 4], [2, 3, 4, 5]),
            {Spin.up: [0, 1, 2, 3], Spin.down: [1, 2, 3, 4]},
            id="list up-down",
        ),
        pytest.param("100:400:4:1", pytest.raises(ValueError), id="error"),
    ],
)
def test_parse_ibands(value, expected):
    if not isinstance(expected, dict):
        with expected:
            parse_ibands(value)
    else:
        parsed = parse_ibands(value)
        parsed = {s: i.tolist() for s, i in parsed.items()}
        assert parsed == expected



"""
Module defining utility functions.
"""
import collections
import copy
import logging
import sys
from multiprocessing.sharedctypes import RawArray
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

import numpy as np
from tqdm.auto import tqdm

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"


logger = logging.getLogger(__name__)

_bar_format = "{desc} {percentage:3.0f}%|{bar}| {elapsed}<{remaining}{postfix}"


def validate_settings(user_settings: Dict[str, Any]) -> Dict[str, Any]:
    """Parse, validate and fill amset settings.

    Missing settings will be inferred from the amset defaults.

    Args:
        user_settings: A dictionary of settings.

    Returns:
        The validated settings.
    """
    from amset.constants import defaults

    settings = copy.deepcopy(defaults)
    settings.update(user_settings)

    # validate the type of some settings
    if isinstance(settings["doping"], (int, float)):
        settings["doping"] = [settings["doping"]]
    elif isinstance(settings["doping"], str):
        settings["doping"] = parse_doping(settings["doping"])

    if isinstance(settings["temperatures"], (int, float)):
        settings["temperatures"] = [settings["temperatures"]]
    elif isinstance(settings["temperatures"], str):
        settings["temperatures"] = parse_temperatures(settings["temperatures"])

    if isinstance(settings["deformation_potential"], str):
        settings["deformation_potential"] = parse_deformation_potential(
            settings["deformation_potential"]
        )
    elif isinstance(settings["deformation_potential"], list):
        settings["deformation_potential"] = tuple(settings["deformation_potential"])

    if settings["static_dielectric"] is not None:
        settings["static_dielectric"] = cast_tensor(settings["static_dielectric"])

    if settings["high_frequency_dielectric"] is not None:
        settings["high_frequency_dielectric"] = cast_tensor(
            settings["high_frequency_dielectric"]
        )

    if settings["elastic_constant"] is not None:
        settings["elastic_constant"] = cast_elastic_tensor(settings["elastic_constant"])

    if settings["piezoelectric_constant"] is not None:
        settings["piezoelectric_constant"] = cast_piezoelectric_tensor(
            settings["piezoelectric_constant"]
        )

    settings["doping"] = np.asarray(settings["doping"], dtype=np.float)
    settings["temperatures"] = np.asarray(settings["temperatures"])

    for charge_setting in ("donor_charge", "acceptor_charge"):
        if charge_setting in settings:
            logger.warning(
                f"The {charge_setting} option has been renamed to defect_charge and "
                "will be removed in June 2021. Please see the documentation for more "
                "details."
            )
            settings["defect_charge"] = settings.pop(charge_setting)

    for setting in settings:
        if setting not in defaults:
            raise ValueError(f"Unrecognised setting: {setting}")

    return settings


def cast_tensor(
    tensor: Union[float, List[float], List[List[float]], np.ndarray]
) -> np.ndarray:
    """Cast a number/list into a 3x3 tensor.

    Args:
        tensor: A number, 3x1 list of numbers, or 3x3 list of numbers.

    Returns:
        A 3x3 tensor.
    """
    from amset.constants import numeric_types

    if isinstance(tensor, numeric_types):
        return np.eye(3) * tensor

    tensor = np.asarray(tensor)
    if len(tensor.shape) == 1:
        return np.diag(tensor)

    if tensor.shape != (3, 3):
        raise ValueError("Unsupported tensor shape.")

    return tensor


def cast_elastic_tensor(
    elastic_tensor: Union[int, float, List[List[float]], np.ndarray]
) -> np.ndarray:
    """Cast elastic tensor from single value or Voigt to full 3x3x3x3 tensor.

    Args:
        elastic_tensor: A single number, 6x6 Voigt tensor, or 3x3x3x3 tensor.

    Returns:
        The elastic constant as a 3x3x3x3 tensor.
    """
    from pymatgen.core.tensors import Tensor

    from amset.constants import numeric_types

    if isinstance(elastic_tensor, numeric_types):
        elastic_tensor = np.eye(6) * elastic_tensor
        elastic_tensor[([3, 4, 5], [3, 4, 5])] /= 2

    elastic_tensor = np.array(elastic_tensor)
    if elastic_tensor.shape == (6, 6):
        elastic_tensor = Tensor.from_voigt(elastic_tensor)

    if elastic_tensor.shape != (3, 3, 3, 3):
        raise ValueError(
            "Unsupported elastic tensor shape. Should be (6, 6) or (3, 3, 3, 3)."
        )

    return np.array(elastic_tensor)


def cast_piezoelectric_tensor(
    piezoelectric_tensor: Union[np.ndarray, List[List[float]], np.ndarray]
) -> np.ndarray:
    """Cast piezoelectric tensor from Voigt form to full 3x3x3 tensor.

    Args:
        piezoelectric_tensor: A 3x6 Voigt tensor, or 3x3x3 tensor.

    Returns:
        The piezoelectric constant as a 3x3x3 tensor.
    """
    from pymatgen.core.tensors import Tensor

    piezoelectric_tensor = np.array(piezoelectric_tensor)
    if piezoelectric_tensor.shape == (3, 6):
        piezoelectric_tensor = Tensor.from_voigt(piezoelectric_tensor)

    if piezoelectric_tensor.shape != (3, 3, 3):
        raise ValueError(
            "Unsupported piezoelectric tensor shape. Should be (3, 6) or (3, 3, 3)."
        )

    return np.array(piezoelectric_tensor)


def tensor_average(tensor: Union[List, np.ndarray]) -> Union[float, np.ndarray]:
    """Calculate the average of the tensor eigenvalues.

    Args:
        tensor: A tensor

    Returns:
        The average of the eigenvalues.
    """
    return np.average(np.linalg.eigvalsh(tensor), axis=-1)


def groupby(
    elements: Union[List[Any], np.ndarray], groups: Union[List[int], np.ndarray]
) -> np.ndarray:
    """Groups elements based on the group indices.

    I.e., if elements is `["a", "b", "1", "2", "c", "d"]` and groups is
    `[2, 0, 1, 2, 0, 0]` the output will be `[["b", "c", "d"], ["1"], ["a", "2"]]`.

    Args:
        elements: A list of objects.
        groups: The groups that the objects belong to.

    Returns:
        A nested list of elements that have been grouped.
    """
    groups = np.array(groups)
    elements = np.array(elements)

    # Get argsort indices, to be used to sort a and b in the next steps
    sidx = groups.argsort(kind="mergesort")
    a_sorted = elements[sidx]
    b_sorted = groups[sidx]

    # Get the group limit indices (start, stop of groups)
    cut_idx = np.flatnonzero(np.r_[True, b_sorted[1:] != b_sorted[:-1], True])

    # Split input array with those start, stop ones
    out = np.array(
        [a_sorted[i:j] for i, j in zip(cut_idx[:-1], cut_idx[1:])], dtype=object
    )
    return out


def cast_dict_list(d):
    """Recursively cast numpy arrays in a dictionary to lists.

    Also casts pymatgen Spin objects to str.

    Args:
        d: A dictionary.

    Returns:
        The casted dictionary.
    """
    from pymatgen.electronic_structure.core import Spin

    if d is None:
        return d

    new_d = {}
    for k, v in d.items():
        # cast keys
        if isinstance(k, Spin):
            k = k.name

        if isinstance(v, collections.abc.Mapping):
            new_d[k] = cast_dict_list(v)
        else:
            # cast values
            if isinstance(v, np.ndarray):
                v = v.tolist()
            elif isinstance(v, tuple):
                v = list(v)

            new_d[k] = v
    return new_d


def cast_dict_ndarray(d):
    """Recursively cast lists in dictionaries to numpy arrays.

    Also casts the keys "up" and "down" to pymatgen Spin objects.

    Args:
        d: A dictionary.

    Returns:
        The casted dictionary.
    """
    from pymatgen.electronic_structure.core import Spin

    if d is None:
        return d

    new_d = {}
    for k, v in d.items():
        # cast keys back to spin
        if isinstance(k, str) and k in ["up", "down"]:
            k = Spin.up if "k" == "up" else Spin.up

        if isinstance(v, collections.abc.Mapping):
            new_d[k] = cast_dict_ndarray(v)
        else:
            # cast values
            if isinstance(v, list):
                v = np.array(v)

            new_d[k] = v
    return new_d


def parse_doping(doping_str: str) -> np.ndarray:
    """Parse doping string.

    Args:
        doping_str: String specifying doping. Can be a list of comma separated numbers,
            or a range specification in log space with start:stop:step
            (i.e., "1e16:1e19:4" would give `[1e16, 1e17, 1e18, 1e19]`).

    Returns:
        The doping as a numpy array.
    """
    doping_str = doping_str.strip().replace(" ", "")

    try:
        if ":" in doping_str:
            parts = list(map(float, doping_str.split(":")))

            if len(parts) != 3:
                raise ValueError

            return np.geomspace(parts[0], parts[1], int(parts[2]))

        else:
            return np.array(list(map(float, doping_str.split(","))))

    except ValueError:
        raise ValueError(f"ERROR: Unrecognised doping format: {doping_str}")


def parse_temperatures(temperatures_str: str) -> np.ndarray:
    """Parse temperature string.

    Args:
        temperatures_str: String specifying temperatures. Can be a list of comma
            separated numbers, or a range specification in with start:stop:step
            (i.e., "100:400:4" would give `[100, 200, 300, 400]`).

    Returns:
        The temperatures as a numpy array.
    """
    temperatures_str = temperatures_str.strip().replace(" ", "")

    try:
        if ":" in temperatures_str:
            parts = list(map(float, temperatures_str.split(":")))

            if len(parts) != 3:
                raise ValueError

            return np.linspace(parts[0], parts[1], int(parts[2]))

        else:
            return np.array(list(map(float, temperatures_str.split(","))))

    except ValueError:
        raise ValueError(f"ERROR: Unrecognised temperature format: {temperatures_str}")


def parse_deformation_potential(
    deformation_pot_str: str,
) -> Union[str, float, Tuple[float, ...]]:
    """Parse deformation potential string.

    Args:
        deformation_pot_str: The deformation potential string. Can be a path to a
            deformation.h5 file, a single deformation potential to use for all bands
            or two deformation potentials separated by a comma for valence and
            conduction bands.

    Returns:
        The deformation potential(s) or path to deformation potential file.
    """

    if "h5" in deformation_pot_str:
        return deformation_pot_str

    deformation_pot_str = deformation_pot_str.strip().replace(" ", "")

    try:
        parts = list(map(float, deformation_pot_str.split(",")))

        if len(parts) == 1:
            return parts[0]
        elif len(parts) == 2:
            return tuple(parts)
        else:
            raise ValueError

    except ValueError:
        raise ValueError(
            "ERROR: Unrecognised deformation potential format: "
            "{}".format(deformation_pot_str)
        )


def get_progress_bar(
    iterable: Optional[Iterable] = None,
    total: Optional[int] = None,
    desc: str = "",
    min_desc_width: int = 18,
    prepend_pipe: bool = True,
) -> tqdm:
    """Get a formatted progress bar.

    One of `iterable` or `total` must be specific.

    Args:
        iterable: An iterable (list, tuple, etc).
        total: The total number of items in the progress bar.
        desc: The descriptive label.
        min_desc_width: Minimum description width.
        prepend_pipe: Add indent and fancy pipe symbol before description.

    Returns:
        The progress bar.
    """
    from amset.constants import output_width

    if prepend_pipe:
        desc = "    ├── " + desc

    desc += ":"

    if len(desc) < min_desc_width:
        desc += " " * (min_desc_width - len(desc))

    if iterable is not None:
        return tqdm(
            iterable=iterable,
            total=total,
            ncols=output_width,
            desc=desc,
            bar_format=_bar_format,
            file=sys.stdout,
        )
    elif total is not None:
        return tqdm(
            total=total,
            ncols=output_width,
            desc=desc,
            bar_format=_bar_format,
            file=sys.stdout,
        )
    else:
        raise ValueError("Error creating progress bar, need total or iterable")


def parse_ibands(ibands: Union[str, Tuple[List[int], List[int]]]) -> Dict:
    """
    Parse ibands string or list to dictionary with Spin.up and Spin.down keys.

    Args:
        ibands: The ibands specification. Can be either a string or tuple. Allowed
            str formats include (i) a series of comma separated numbers, i.e., "1, 2, 3"
            that results in an ibands of `{Spin.up: [0, 1, 2]}` (bands in pymatgen are
            zero indexed; (ii) a range can be specified, i.e., "1:3" results
            in the same ibands as above (iii) Spin up and spin down ibands (only useful
            for spin-polarized systems, can be specified by separating the spin channels
            with a period. I.e., "1,2,3.4:6", results in an ibands of `{Spin.up:
            [0, 1, 2], Spin.down: [3, 4, 5]}`. Alternatively, a tuple of lists can be
            given. I.e., `([1, 2, 3], [4, 5, 6])` gives the same ibands as the last
            example.

    Returns:
        The ibands specification.
    """
    from pymatgen.electronic_structure.core import Spin

    new_ibands = {}
    if isinstance(ibands, str):
        try:
            for spin, spin_ibands in zip((Spin.up, Spin.down), ibands.split(".")):
                if ":" in spin_ibands:
                    parts = list(map(int, spin_ibands.split(":")))
                    if len(parts) != 2:
                        raise ValueError
                    new_ibands[spin] = list(range(parts[0], parts[1] + 1))
                else:
                    new_ibands[spin] = list(map(int, spin_ibands.split(",")))

        except ValueError:
            raise ValueError(f"ERROR: Unrecognised ibands format: {ibands}")
    elif isinstance(ibands, (list, tuple)):
        if not isinstance(ibands[0], (list, tuple)):
            new_ibands[Spin.up] = ibands
        else:
            new_ibands[Spin.up] = ibands[0]
            new_ibands[Spin.down] = ibands[1]
    return {s: np.array(i, dtype=int) - 1 for s, i in new_ibands.items()}


def create_shared_array(data: np.ndarray, return_shared_data=False):
    data = np.asarray(data)
    if data.dtype == np.complex:
        data_type = "complex"
        data_buffer = RawArray("d", int(np.prod(data.shape)) * 2)
    else:
        data_type = np.ctypeslib.as_ctypes_type(data.dtype)
        data_buffer = RawArray(data_type, int(np.prod(data.shape)))

    buffer = (data_buffer, data.shape, data_type)
    data_shared = array_from_buffer(buffer)
    data_shared[:] = data[:]

    if return_shared_data:
        return buffer, data_shared
    else:
        return buffer


def create_shared_dict_array(data: Dict[Any, np.ndarray], return_shared_data=False):
    # turns a dict of key: np.ndarray to a dict of key: buffer
    data_buffer = {}
    data_shared = {}
    for key, value in data.items():
        data_buffer[key], data_shared[key] = create_shared_array(
            value, return_shared_data=True
        )

    if return_shared_data:
        return data_buffer, data_shared
    else:
        return data_buffer


def array_from_buffer(buffer):
    data_buffer, data_shape, data_type = buffer
    if data_type == "complex":
        return np.frombuffer(data_buffer).view(np.complex).reshape(data_shape)
    else:
        return np.frombuffer(data_buffer, dtype=data_type).reshape(data_shape)


def dict_array_from_buffer(buffer):
    data = {}
    for key, value_buffer in buffer.items():
        data[key] = array_from_buffer(value_buffer)
    return data



from pathlib import Path

import click
import numpy as np

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"

from amset.tools.common import zero_weighted_type
from amset.util import parse_ibands


@click.command()
@click.option("-w", "--wavecar", default="WAVECAR", help="WAVECAR file")
@click.option("-v", "--vasprun", default="vasprun.xml", help="vasprun.xml file")
@click.option(
    "-p", "--potcar", default="POTCAR", help="POTCAR (only needed with --pawpyseed)"
)
@click.option("-d", "--directory", help="directory to look for files")
@click.option(
    "-e", "--energy-cutoff", type=float, help="energy cutoff for finding bands"
)
@click.option(
    "-b",
    "--bands",
    type=str,
    help="bands at which to extract the wavefunction, e.g., '1:10' (overrides "
    "energy-cutoff)",
)
@click.option(
    "-z",
    "--zero-weighted-kpoints",
    help="how to process zero-weighted k-points",
    type=zero_weighted_type,
)
@click.option(
    "-c",
    "--planewave-cutoff",
    default=None,
    type=float,
    help="planewave cutoff for coefficients",
)
@click.option(
    "--pawpyseed", is_flag=True, help="use pawpyseed to generate coefficients"
)
@click.option("-o", "--output", default="wavefunction.h5", help="output file path")
@click.option(
    "-v", "--vasp-type", help="set the vasp wavefunction type (std, ncl, or gam)"
)
def wave(**kwargs):
    """Extract wavefunction coefficients from a WAVECAR"""
    from pymatgen.io.vasp import BSVasprun

    from amset.constants import defaults
    from amset.electronic_structure.common import (
        get_band_structure,
        get_ibands,
        get_zero_weighted_kpoint_indices,
    )
    from amset.tools.common import echo_ibands
    from amset.wavefunction.io import write_coefficients

    output = kwargs.pop("output")
    planewave_cutoff = kwargs.pop("planewave_cutoff")
    pawpyseed = kwargs.pop("pawpyseed")

    energy_cutoff = kwargs.pop("energy_cutoff")
    if not energy_cutoff:
        energy_cutoff = defaults["energy_cutoff"]

    if kwargs["directory"]:
        vasprun_file = Path(kwargs["directory"]) / "vasprun.xml"
    else:
        vasprun_file = kwargs["vasprun"]

    try:
        vr = BSVasprun(vasprun_file)
    except FileNotFoundError:
        vr = BSVasprun(str(vasprun_file) + ".gz")

    zwk_mode = kwargs.pop("zero_weighted_kpoints")
    if not zwk_mode:
        zwk_mode = defaults["zero_weighted_kpoints"]

    bs = get_band_structure(vr, zero_weighted=zwk_mode)

    if "bands" in kwargs and kwargs["bands"] is not None:
        ibands = parse_ibands(kwargs["bands"])
    else:
        ibands = get_ibands(energy_cutoff, bs)
    ikpoints = get_zero_weighted_kpoint_indices(vr, zwk_mode)

    click.echo("******* Getting wavefunction coefficients *******\n")
    echo_ibands(ibands, bs.is_spin_polarized)
    click.echo("")

    if pawpyseed:
        coeffs, gpoints, kpoints = _wavefunction_pawpy(
            bs, ibands, planewave_cutoff, ikpoints, **kwargs
        )
    else:
        coeffs, gpoints = _wavefunction_vasp(
            ibands, planewave_cutoff, ikpoints, **kwargs
        )
        kpoints = np.array([k.frac_coords for k in bs.kpoints])

    structure = vr.final_structure

    click.echo(f"Writing coefficients to {output}")
    write_coefficients(coeffs, gpoints, kpoints, structure, filename=output)


def _wavefunction_vasp(ibands, planewave_cutoff, ikpoints, **kwargs):
    from amset.wavefunction.vasp import (
        get_converged_encut,
        get_wavefunction,
        get_wavefunction_coefficients,
    )

    directory = kwargs.pop("directory")
    wavecar = kwargs.pop("wavecar")
    vasp_type = kwargs.pop("vasp_type")
    wf = get_wavefunction(wavecar=wavecar, directory=directory, vasp_type=vasp_type)

    if not planewave_cutoff:
        click.echo("******* Automatically choosing plane wave cutoff *******")
        planewave_cutoff = get_converged_encut(
            wf, iband=ibands, ikpoints=ikpoints, n_samples=2000, std_tol=0.02
        )
        click.echo(f"\nUsing cutoff: {planewave_cutoff} eV")

    return get_wavefunction_coefficients(
        wf, iband=ibands, ikpoints=ikpoints, encut=planewave_cutoff
    )


def _wavefunction_pawpy(bs, ibands, planewave_cutoff, ikpoints, **pawpy_kwargs):
    from amset.wavefunction.pawpyseed import (
        get_converged_encut,
        get_wavefunction,
        get_wavefunction_coefficients,
    )

    wf = get_wavefunction(**pawpy_kwargs)
    if not planewave_cutoff:
        click.echo("******* Automatically choosing plane wave cutoff *******")
        planewave_cutoff = get_converged_encut(
            wf,
            bs,
            iband=ibands,
            ikpoints=ikpoints,
            max_encut=600,
            n_samples=2000,
            std_tol=0.02,
        )
        click.echo(f"\nUsing cutoff: {planewave_cutoff} eV")

    return get_wavefunction_coefficients(
        wf, bs, iband=ibands, ikpoints=ikpoints, encut=planewave_cutoff
    )



# import os
# import unittest
#
# import numpy as np
# from numpy.testing import assert_array_equal
# from pymatgen.electronic_structure.core import Spin
# from pymatgen.io.vasp import Vasprun
#
# from amset.interpolation.bandstructure import Interpolator
# from amset.log import initialize_amset_logger
#
# test_dir = os.path.dirname(os.path.abspath(__file__))
# gaas_files = os.path.join(test_dir, "..", "..", "..", "examples", "GaAs")
# tin_dioxide_files = os.path.join(test_dir, "..", "..", "..", "examples", "SnO2")
# si_files = os.path.join(test_dir, "..", "..", "..", "examples", "Si")
# pbs_files = os.path.join(test_dir, "..", "..", "..", "examples", "PbS")
#
#
# @unittest.skip("Outdated tests")
# class TestBoltzTraP2Interpolater(unittest.TestCase):
#     """Tests for interpolating a band structure using BoltzTraP2."""
#
#     def setUp(self):
#         vr = Vasprun(
#             os.path.join(gaas_files, "vasprun.xml.gz"), parse_projected_eigen=True
#         )
#         bs = vr.get_band_structure()
#         num_electrons = vr.parameters["NELECT"]
#
#         self.kpoints = np.array(vr.actual_kpoints)
#         self.interpolater = Interpolator(
#             bs, num_electrons, interpolate_projections=True, interpolation_factor=1
#         )
#
#     def test_initialisation(self):
#         """Test coefficients and parameters are calculated correctly."""
#         self.interpolater.initialize()
#         params = self.interpolater._parameters
#         assert_array_equal(params[0][0], [[0, 0, 0]])
#         self.assertAlmostEqual(params[1][0][0], 9.14055614)
#         self.assertAlmostEqual(params[2][0][0].real, -2.33144546)
#
#     def test_get_energies(self):
#         """Test getting the interpolated energy, velocity and effective mass."""
#
#         # test just getting energy
#         energies = self.interpolater.get_energies(
#             self.kpoints, 25, return_velocity=False, curvature=False
#         )
#         self.assertEqual(energies.shape, (138,))
#         self.assertAlmostEqual(energies[0], 3.852399483908641)
#
#         # test energy + velocity
#         energies, velocities = self.interpolater.get_energies(
#             self.kpoints, 25, return_velocity=True, curvature=False
#         )
#         self.assertEqual(energies.shape, (138,))
#         self.assertAlmostEqual(energies[0], 3.852399483908641)
#         self.assertEqual(velocities.shape, (138, 3))
#         self.assertAlmostEqual(velocities[10][0], 5.401166156893334e6, places=0)
#
#         # test energy + effective_mass
#         energies, effective_masses = self.interpolater.get_energies(
#             self.kpoints, 25, return_velocity=False, curvature=True
#         )
#         self.assertEqual(energies.shape, (138,))
#         self.assertAlmostEqual(energies[0], 3.852399483908641)
#         self.assertEqual(effective_masses.shape, (138, 3, 3))
#         self.assertAlmostEqual(effective_masses[10][0][0], 0.2242569613626494)
#
#         # test energy + velocity + effective_mass
#         energies, velocities, effective_masses = self.interpolater.get_energies(
#             self.kpoints, 25, return_velocity=True, curvature=True
#         )
#         self.assertEqual(energies.shape, (138,))
#         self.assertAlmostEqual(energies[0], 3.852399483908641)
#         self.assertEqual(velocities.shape, (138, 3))
#         self.assertAlmostEqual(velocities[10][0], 5.401166156893334e6, places=0)
#         self.assertEqual(effective_masses.shape, (138, 3, 3))
#         self.assertAlmostEqual(effective_masses[10][0][0], 0.2242569613626494)
#
#     def test_get_energies_scissor(self):
#         """Test scissoring of band energies."""
#
#         # test valence band
#         energies = self.interpolater.get_energies(
#             self.kpoints, 25, return_velocity=False, curvature=False, scissor=1.0
#         )
#         self.assertEqual(energies.shape, (138,))
#         self.assertAlmostEqual(energies[0], 3.852399483908641 - 0.5)
#
#         # test conduction band
#         energies = self.interpolater.get_energies(
#             self.kpoints, 33, return_velocity=False, curvature=False, scissor=1.0
#         )
#         self.assertEqual(energies.shape, (138,))
#         self.assertAlmostEqual(energies[0], 7.301700765 + 0.5)
#
#     def test_get_energies_multiple_bands(self):
#         """Test getting the interpolated data for multiple bands."""
#         # test just getting energy
#         energies = self.interpolater.get_energies(
#             self.kpoints, [25, 35], return_velocity=False, curvature=False
#         )
#         self.assertEqual(energies.shape, (2, 138))
#         self.assertAlmostEqual(energies[0][0], 3.852399483908641)
#         self.assertAlmostEqual(energies[1][0], 9.594401616456384)
#
#         # test energy + velocity
#         energies, velocities = self.interpolater.get_energies(
#             self.kpoints, [25, 35], return_velocity=True, curvature=False
#         )
#         self.assertEqual(energies.shape, (2, 138))
#         self.assertAlmostEqual(energies[0][0], 3.852399483908641)
#         self.assertAlmostEqual(energies[1][0], 9.594401616456384)
#         self.assertEqual(velocities.shape, (2, 138, 3))
#         self.assertAlmostEqual(velocities[0][10][0], 5.401166156893334e6, places=0)
#         self.assertAlmostEqual(velocities[1][10][0], 1.1686720831758736e8, places=0)
#
#         # test energy + effective_mass
#         energies, effective_masses = self.interpolater.get_energies(
#             self.kpoints, [25, 35], return_velocity=False, curvature=True
#         )
#         self.assertEqual(energies.shape, (2, 138))
#         self.assertAlmostEqual(energies[0][0], 3.852399483908641)
#         self.assertAlmostEqual(energies[1][0], 9.594401616456384)
#         self.assertEqual(effective_masses.shape, (2, 138, 3, 3))
#         self.assertAlmostEqual(effective_masses[0][10][0][0], 0.224256961362649)
#         self.assertAlmostEqual(effective_masses[1][10][0][0], 0.009057103344700)
#
#         # test energy + velocity + effective_mass
#         energies, velocities, effective_masses = self.interpolater.get_energies(
#             self.kpoints, [25, 35], return_velocity=True, curvature=True
#         )
#         self.assertEqual(energies.shape, (2, 138))
#         self.assertAlmostEqual(energies[0][0], 3.852399483908641)
#         self.assertAlmostEqual(energies[1][0], 9.594401616456384)
#         self.assertEqual(velocities.shape, (2, 138, 3))
#         self.assertAlmostEqual(velocities[0][10][0], 5.401166156893334e6, places=0)
#         self.assertAlmostEqual(velocities[1][10][0], 1.1686720831758736e8, places=0)
#         self.assertEqual(effective_masses.shape, (2, 138, 3, 3))
#         self.assertAlmostEqual(effective_masses[0][10][0][0], 0.224256961362649)
#         self.assertAlmostEqual(effective_masses[1][10][0][0], 0.009057103344700)
#
#     def test_get_energies_all_bands(self):
#         # test all bands
#         energies, velocities, effective_masses = self.interpolater.get_energies(
#             self.kpoints, None, return_velocity=True, curvature=True
#         )
#         self.assertEqual(energies.shape, (96, 138))
#         self.assertAlmostEqual(energies[25][0], 3.852399483908641)
#         self.assertAlmostEqual(energies[35][0], 9.594401616456384)
#         self.assertEqual(velocities.shape, (96, 138, 3))
#         self.assertAlmostEqual(velocities[25][10][0], 5.401166156893334e6, places=0)
#         self.assertAlmostEqual(velocities[35][10][0], 1.1686720831758736e8, places=0)
#         self.assertEqual(effective_masses.shape, (96, 138, 3, 3))
#         self.assertAlmostEqual(effective_masses[25][10][0][0], 0.22425696136264)
#         self.assertAlmostEqual(effective_masses[35][10][0][0], 0.00905710334470)
#
#     def test_get_dos(self):
#         """Test generating the interpolated DOS."""
#         dos = self.interpolater.get_dos([10, 10, 10], emin=-10, emax=10, width=0.075)
#         self.assertEqual(dos.shape, (20000, 2))
#         self.assertEqual(dos[0][0], -10)
#         self.assertAlmostEqual(dos[15000][1], 3.5362612128412807)
#
#     def test_get_energies_symprec(self):
#         # vr = Vasprun(os.path.join(tin_dioxide_files, 'vasprun.xml.gz'),
#         #              parse_projected_eigen=True)
#         vr = Vasprun(
#             os.path.join(pbs_files, "vasprun.xml.gz"), parse_projected_eigen=True
#         )
#         bs = vr.get_band_structure()
#         num_electrons = vr.parameters["NELECT"]
#         interpolater = Interpolator(
#             bs, num_electrons, interpolate_projections=True, interpolation_factor=1
#         )
#
#         ir_kpoints, weights, kpoints, ir_kpoints_idx, ir_to_full_idx = get_kpoints(
#             [13, 15, 29],
#             vr.final_structure,
#             boltztrap_ordering=True,
#             return_full_kpoints=True,
#         )
#
#         initialize_amset_logger()
#
#         (
#             energies,
#             velocities,
#             curvature,
#             projections,
#             sym_info,
#         ) = interpolater.get_energies(
#             kpoints,
#             None,
#             return_velocity=True,
#             atomic_units=True,
#             return_curvature=True,
#             return_projections=True,
#             symprec=0.1,
#             return_vel_outer_prod=True,
#             return_kpoint_mapping=True,
#         )
#
#         (
#             energies_no_sym,
#             velocities_no_sym,
#             curvature_no_sym,
#             projections_no_sym,
#         ) = interpolater.get_energies(
#             kpoints,
#             None,
#             return_velocity=True,
#             atomic_units=True,
#             return_curvature=True,
#             return_projections=True,
#             return_vel_outer_prod=True,
#             symprec=None,
#         )
#
#         np.testing.assert_array_equal(ir_to_full_idx, sym_info["ir_to_full_idx"])
#         np.testing.assert_array_equal(ir_kpoints_idx, sym_info["ir_kpoints_idx"])
#
#         # print(velocities[Spin.up][5, :, :, -3:])
#         # print(velocities_no_sym[Spin.up][5, :, :, -3:])
#         # print(sym_info["ir_to_full_idx"][-10:])
#
#         np.testing.assert_array_almost_equal(
#             energies[Spin.up], energies_no_sym[Spin.up], decimal=12
#         )
#         np.testing.assert_array_almost_equal(
#             velocities[Spin.up], velocities_no_sym[Spin.up], decimal=12
#         )
#         np.testing.assert_array_almost_equal(
#             curvature[Spin.up], curvature_no_sym[Spin.up], decimal=12
#         )
#
#         for l in projections[Spin.up]:
#             np.testing.assert_array_almost_equal(
#                 projections[Spin.up][l], projections_no_sym[Spin.up][l]
#             )
#
#     def test_get_energies_interpolater(self):
#
#         initialize_amset_logger()
#
#         amset_data = self.interpolater.get_amset_data()
#
#         energies, velocities, projections, sym_info = self.interpolater.get_energies(
#             amset_data.kpoints,
#             None,
#             return_velocity=True,
#             atomic_units=True,
#             curvature=False,
#             return_projections=True,
#             symprec=0.1,
#             return_vel_outer_prod=True,
#             return_kpoint_mapping=True,
#         )
#
#         np.testing.assert_array_almost_equal(
#             energies[Spin.up], amset_data.energies[Spin.up]
#         )
#         np.testing.assert_array_almost_equal(
#             velocities[Spin.up], amset_data.velocities_product[Spin.up]
#         )
#
#         for l in projections[Spin.up]:
#             np.testing.assert_array_almost_equal(
#                 projections[Spin.up][l], amset_data._projections[Spin.up][l]
#             )
#
#         np.testing.assert_array_equal(
#             sym_info["ir_kpoints_idx"], amset_data.ir_kpoints_idx
#         )
#         np.testing.assert_array_equal(
#             sym_info["ir_to_full_idx"], amset_data.ir_to_full_kpoint_mapping
#         )



import logging

import numpy as np
from pymatgen.analysis.elasticity.strain import Deformation
from pymatgen.core.tensors import TensorMapping

from amset.constants import defaults
from amset.electronic_structure.kpoints import (
    get_kpoint_indices,
    get_kpoints_from_bandstructure,
    get_mesh_from_kpoint_diff,
    kpoints_to_first_bz,
)
from amset.electronic_structure.symmetry import (
    expand_bandstructure,
    get_symmops,
    rotate_bandstructure,
    similarity_transformation,
)

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"


_mapping_tol = 0.002

logger = logging.getLogger(__name__)


def get_mesh_from_band_structure(bandstructure):
    kpoints = np.array([k.frac_coords for k in bandstructure.kpoints])
    mesh, is_shifted = get_mesh_from_kpoint_diff(kpoints)
    return tuple(mesh.round().astype(int)), is_shifted


def calculate_deformation(bulk_structure, deformed_structure):
    """
    Args:
        bulk_structure: undeformed structure
        deformed_structure: deformed structure

    Returns:
        deformation matrix
    """
    bulk_lattice = bulk_structure.lattice.matrix
    deformed_lattice = deformed_structure.lattice.matrix
    return np.transpose(np.dot(np.linalg.inv(bulk_lattice), deformed_lattice)).round(10)


def get_strain_mapping(bulk_structure, deformation_calculations):
    strain_mapping = TensorMapping(tol=_mapping_tol)
    for i, calc in enumerate(deformation_calculations):
        deformed_structure = calc["bandstructure"].structure
        matrix = calculate_deformation(bulk_structure, deformed_structure)
        strain = Deformation(matrix).green_lagrange_strain
        strain_mapping[strain] = calc
    return strain_mapping


def get_symmetrized_strain_mapping(
    bulk_structure,
    strain_mapping,
    symprec=defaults["symprec"],
    symprec_deformation=defaults["symprec"] / 100,
):
    # get symmetry operations of the bulk structure
    frac_ops = get_symmops(bulk_structure, symprec=symprec)

    for strain, calc in strain_mapping.items():
        # expand band structure to cover full brillouin zone, otherwise rotation won't
        # include all necessary points
        # default symprec is lower otherwise the strain will not be noticed
        calc["bandstructure"] = expand_bandstructure(
            calc["bandstructure"], symprec=symprec_deformation
        )

    for strain, calc in strain_mapping.items():
        k_old = get_kpoints_from_bandstructure(calc["bandstructure"], sort=True)
        k_old = kpoints_to_first_bz(k_old)

        for frac_op in frac_ops:
            # apply cartesian transformation matrix from the right side
            # hence the transpose
            r_cart = similarity_transformation(
                bulk_structure.lattice.matrix.T, frac_op.rotation_matrix.T
            )
            tstrain = strain.rotate(r_cart)

            independent = tstrain.get_deformation_matrix().is_independent(_mapping_tol)
            if independent and tstrain not in strain_mapping:
                rband = rotate_bandstructure(calc["bandstructure"], frac_op)

                k_new = get_kpoints_from_bandstructure(rband, sort=True)
                k_new = kpoints_to_first_bz(k_new)

                # check whether k-points match; if no match found this indicates
                # that the real and reciprocal lattice have different symmetries
                kpoints_match = np.max(np.linalg.norm(k_old - k_new, axis=1)) < 0.001

                if kpoints_match:
                    tcalc = {"reference": calc["reference"], "bandstructure": rband}
                    strain_mapping[tstrain] = tcalc

    return strain_mapping


def get_strain_deformation_potential(
    strain,
    bulk_bandstructure,
    deformation_bandstructure,
    bulk_reference,
    deformation_reference,
):
    strain = strain.round(5)
    flat_strain = strain.ravel()
    strain_amount = flat_strain[np.abs(flat_strain).argmax()]
    ref_diff = bulk_reference - deformation_reference

    kpoints = get_kpoints_from_bandstructure(bulk_bandstructure)
    mesh, is_shifted = get_mesh_from_band_structure(bulk_bandstructure)
    indices_to_keep = get_kpoint_indices(kpoints, mesh, is_shifted=is_shifted)

    deform_kpoints = get_kpoints_from_bandstructure(deformation_bandstructure)
    deform_indices = get_kpoint_indices(deform_kpoints, mesh, is_shifted=is_shifted)

    if not set(indices_to_keep).issubset(set(deform_indices)):
        raise RuntimeError(
            "Deformation band structure doesn't contain the same k-points "
            "as the bulk band structure. Try changing symprec."
        )

    deform_map = np.full(np.max(deform_indices) + 1, -1)
    deform_map[deform_indices] = np.arange(len(deform_indices))
    select_indices = deform_map[indices_to_keep]

    energy_diff = {}
    for spin, spin_origin in bulk_bandstructure.bands.items():
        diff = spin_origin - deformation_bandstructure.bands[spin][:, select_indices]
        diff -= ref_diff
        energy_diff[spin] = np.abs(diff / strain_amount)

    return energy_diff


def calculate_deformation_potentials(bulk_calculation, strain_mapping):
    deformation_potentials = {
        s: np.zeros(b.shape + (3, 3))
        for s, b in bulk_calculation["bandstructure"].bands.items()
    }
    norm = np.zeros((3, 3))
    for strain, deformation_calculation in strain_mapping.items():
        deform = get_strain_deformation_potential(
            strain,
            bulk_calculation["bandstructure"],
            deformation_calculation["bandstructure"],
            bulk_calculation["reference"],
            deformation_calculation["reference"],
        )

        # sometimes the strain includes numerical noise, this will filter out components
        # of the strain that are noisy. In reality there should only be one or two
        # components of the strain and they should have the same magnitude
        max_strain = np.abs(strain).max()
        strain_loc = np.abs(strain) > 0.25 * max_strain
        loc_x, loc_y = np.where(strain_loc)
        for spin, spin_deform in deform.items():
            deformation_potentials[spin][:, :, loc_x, loc_y] += spin_deform[..., None]
            norm += strain_loc

    for spin in deformation_potentials:
        deformation_potentials[spin] /= norm[None, None]

    return deformation_potentials


def strain_coverage_ok(strains):
    sum_strains = np.abs(strains).sum(axis=0)
    sum_strains = sum_strains.round(5)
    return not np.any(sum_strains == 0)


def extract_bands(deformation_potentials, ibands):
    return {spin: deformation_potentials[spin][b_idx] for spin, b_idx in ibands.items()}
